@@ -44,6 +44,10 @@
 #include <netinet/in.h>
 #include <ifaddrs.h>
 
+#ifdef HAVE_LIBURING
+#include <liburing.h>
+#endif
+
 #include <ofi_osd.h>
 #include <ofi_list.h>
 
@@ -101,6 +105,11 @@ static inline uint64_t ntohll(uint64_t x) { return x; }
 #endif
 
 
+enum {
+	OFI_EINPROGRESS_ASYNC = 512,	/* Async sockets */
+	OFI_EINPROGRESS_URING = 513,	/* io_uring */
+};
+
 static inline int ofi_recvall_socket(SOCKET sock, void *buf, size_t len)
 {
 	ssize_t ret;
@@ -123,7 +132,297 @@ static inline int ofi_sendall_socket(SOCKET sock, const void *buf, size_t len)
 	return (size_t) sent != len;
 }
 
-int ofi_discard_socket(SOCKET sock, size_t len);
+ssize_t ofi_discard_socket(SOCKET sock, size_t len);
+
+/*
+ * Socket API
+ */
+#ifdef HAVE_LIBURING
+typedef struct io_uring ofi_io_uring_t;
+typedef struct io_uring_cqe ofi_io_uring_cqe_t;
+#else
+typedef int ofi_io_uring_t;
+typedef struct {
+	uint64_t user_data;
+	int32_t res;
+	uint32_t flags;
+} ofi_io_uring_cqe_t;
+#endif
+
+struct ofi_sockctx {
+	void *context;
+	bool uring_sqe_inuse;
+};
+
+struct ofi_sockapi_uring {
+	ofi_io_uring_t *io_uring;
+	uint64_t credits;
+};
+
+struct ofi_sockapi {
+	struct ofi_sockapi_uring tx_uring;
+	struct ofi_sockapi_uring rx_uring;
+
+	int (*connect)(struct ofi_sockapi *sockapi, SOCKET sock,
+		       const struct sockaddr *addr, socklen_t addrlen,
+		       struct ofi_sockctx *ctx);
+	int (*accept)(struct ofi_sockapi *sockapi, SOCKET sock,
+		      struct sockaddr *addr, socklen_t *addrlen,
+		      struct ofi_sockctx *ctx);
+
+	ssize_t (*send)(struct ofi_sockapi *sockapi, SOCKET sock, const void *buf,
+			size_t len, int flags, struct ofi_sockctx *ctx);
+	ssize_t (*sendv)(struct ofi_sockapi *sockapi, SOCKET sock,
+			 const struct iovec *iov, size_t cnt, int flags,
+			 struct ofi_sockctx *ctx);
+	ssize_t (*recv)(struct ofi_sockapi *sockapi, SOCKET sock, void *buf,
+			size_t len, int flags, struct ofi_sockctx *ctx);
+	ssize_t (*recvv)(struct ofi_sockapi *sockapi, SOCKET sock,
+			 struct iovec *iov, size_t cnt, int flags,
+			 struct ofi_sockctx *ctx);
+};
+
+static inline void
+ofi_sockctx_init(struct ofi_sockctx *sockctx, void *context)
+{
+	sockctx->context = context;
+	sockctx->uring_sqe_inuse = false;
+}
+
+static inline int
+ofi_sockapi_connect_socket(struct ofi_sockapi *sockapi, SOCKET sock,
+			   const struct sockaddr *addr, socklen_t addrlen,
+			   struct ofi_sockctx *ctx)
+{
+	int ret;
+
+	OFI_UNUSED(sockapi);
+	OFI_UNUSED(ctx);
+
+	ret = connect(sock, addr, addrlen);
+	if (ret < 0)
+		return -ofi_sockerr();
+	return ret;
+}
+
+static inline int
+ofi_sockapi_accept_socket(struct ofi_sockapi *sockapi, SOCKET sock,
+			  struct sockaddr *addr, socklen_t *addrlen,
+			  struct ofi_sockctx *ctx)
+{
+	int ret;
+
+	OFI_UNUSED(sockapi);
+	OFI_UNUSED(ctx);
+
+	ret = accept(sock, addr, addrlen);
+	if (ret < 0)
+		return -ofi_sockerr();
+	return ret;
+}
+
+static inline ssize_t
+ofi_sockapi_send_socket(struct ofi_sockapi *sockapi, SOCKET sock, const void *buf,
+			size_t len, int flags, struct ofi_sockctx *ctx)
+{
+	ssize_t ret;
+	int err;
+
+	OFI_UNUSED(sockapi);
+	OFI_UNUSED(ctx);
+	ret = ofi_send_socket(sock, buf, len, flags);
+	if (ret < 0) {
+		err = ofi_sockerr();
+		if (err == EPIPE)
+			return -FI_ENOTCONN;
+		if (err == EWOULDBLOCK)
+			return -FI_EAGAIN;
+		return -err;
+	}
+	return ret;
+}
+
+static inline ssize_t
+ofi_sockapi_sendv_socket(struct ofi_sockapi *sockapi, SOCKET sock,
+			 const struct iovec *iov, size_t cnt, int flags,
+			 struct ofi_sockctx *ctx)
+{
+	ssize_t ret;
+	int err;
+
+	OFI_UNUSED(sockapi);
+	OFI_UNUSED(ctx);
+	ret = ofi_sendv_socket(sock, iov, cnt, flags);
+	if (ret < 0) {
+		err = ofi_sockerr();
+		if (err == EPIPE)
+			return -FI_ENOTCONN;
+		if (err == EWOULDBLOCK)
+			return -FI_EAGAIN;
+		return -err;
+	}
+	return ret;
+}
+
+static inline ssize_t
+ofi_sockapi_recv_socket(struct ofi_sockapi *sockapi, SOCKET sock, void *buf,
+			size_t len, int flags, struct ofi_sockctx *ctx)
+{
+	ssize_t ret;
+
+	OFI_UNUSED(sockapi);
+	OFI_UNUSED(ctx);
+	ret = ofi_recv_socket(sock, buf, len, flags);
+	if (ret <= 0)
+		return ret ? -ofi_sockerr() : -FI_ENOTCONN;
+	return ret;
+}
+
+static inline ssize_t
+ofi_sockapi_recvv_socket(struct ofi_sockapi *sockapi, SOCKET sock,
+			 struct iovec *iov, size_t cnt, int flags,
+			 struct ofi_sockctx *ctx)
+{
+	ssize_t ret;
+
+	OFI_UNUSED(sockapi);
+	OFI_UNUSED(ctx);
+	ret = ofi_recvv_socket(sock, iov, cnt, flags);
+	if (ret <= 0)
+		return ret ? -ofi_sockerr() : -FI_ENOTCONN;
+	return ret;
+}
+
+#ifdef HAVE_LIBURING
+int ofi_sockapi_connect_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+			      const struct sockaddr *addr, socklen_t addrlen,
+			      struct ofi_sockctx *ctx);
+int ofi_sockapi_accept_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+			     struct sockaddr *addr, socklen_t *addrlen,
+			     struct ofi_sockctx *ctx);
+
+ssize_t ofi_sockapi_send_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+			       const void *buf, size_t len, int flags,
+			       struct ofi_sockctx *ctx);
+ssize_t ofi_sockapi_sendv_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+				const struct iovec *iov, size_t cnt, int flags,
+				struct ofi_sockctx *ctx);
+ssize_t ofi_sockapi_recv_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+			       void *buf, size_t len, int flags,
+			       struct ofi_sockctx *ctx);
+ssize_t ofi_sockapi_recvv_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+				struct iovec *iov, size_t cnt, int flags,
+				struct ofi_sockctx *ctx);
+
+int ofi_sockctx_uring_cancel(struct ofi_sockapi_uring *uring,
+			     struct ofi_sockctx *canceled_ctx,
+			     struct ofi_sockctx *ctx);
+
+int ofi_uring_init(ofi_io_uring_t *io_uring, size_t entries);
+int ofi_uring_destroy(ofi_io_uring_t *io_uring);
+
+static inline int ofi_uring_get_fd(ofi_io_uring_t *io_uring)
+{
+	return io_uring->ring_fd;
+}
+
+static inline unsigned int ofi_uring_sq_ready(ofi_io_uring_t *io_uring)
+{
+	return io_uring_sq_ready(io_uring);
+}
+
+static inline unsigned int ofi_uring_sq_space_left(ofi_io_uring_t *io_uring)
+{
+	return io_uring_sq_space_left(io_uring);
+}
+
+static inline unsigned int ofi_uring_cq_ready(ofi_io_uring_t *io_uring)
+{
+	return io_uring_cq_ready(io_uring);
+}
+
+static inline int ofi_uring_submit(ofi_io_uring_t *io_uring)
+{
+	return io_uring_submit(io_uring);
+}
+
+static inline unsigned int
+ofi_uring_peek_batch_cqe(ofi_io_uring_t *io_uring,
+			 ofi_io_uring_cqe_t **cqes, unsigned int count)
+{
+	return io_uring_peek_batch_cqe(io_uring, cqes, count);
+}
+
+static inline void ofi_uring_cq_advance(ofi_io_uring_t *io_uring, unsigned int count)
+{
+	io_uring_cq_advance(io_uring, count);
+}
+#else
+static inline int
+ofi_sockapi_connect_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+			  const struct sockaddr *addr, socklen_t addrlen,
+			  struct ofi_sockctx *ctx)
+{
+	return -FI_ENOSYS;
+}
+
+static inline int
+ofi_sockapi_accept_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+			 struct sockaddr *addr, socklen_t *addrlen,
+			 struct ofi_sockctx *ctx)
+{
+	return -FI_ENOSYS;
+}
+
+static inline ssize_t
+ofi_sockapi_send_uring(struct ofi_sockapi *sockapi, SOCKET sock, const void *buf,
+		       size_t len, int flags, struct ofi_sockctx *ctx)
+{
+	return -FI_ENOSYS;
+}
+
+static inline ssize_t
+ofi_sockapi_sendv_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+			const struct iovec *iov, size_t cnt, int flags,
+			struct ofi_sockctx *ctx)
+{
+	return -FI_ENOSYS;
+}
+
+static inline ssize_t
+ofi_sockapi_recv_uring(struct ofi_sockapi *sockapi, SOCKET sock, void *buf,
+		       size_t len, int flags, struct ofi_sockctx *ctx)
+{
+	return -FI_ENOSYS;
+}
+
+static inline ssize_t
+ofi_sockapi_recvv_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+			struct iovec *iov, size_t cnt, int flags,
+			struct ofi_sockctx *ctx)
+{
+	return -FI_ENOSYS;
+}
+
+
+static inline int
+ofi_sockctx_uring_cancel(struct ofi_sockapi_uring *uring,
+			 struct ofi_sockctx *canceled_ctx,
+			 struct ofi_sockctx *ctx)
+{
+	return -FI_ENOSYS;
+}
+
+#define ofi_uring_init(io_uring, entries) -FI_ENOSYS
+#define ofi_uring_destroy(io_uring) -FI_ENOSYS
+#define ofi_uring_get_fd(io_uring) INVALID_SOCKET
+#define ofi_uring_sq_ready(io_uring) 0
+#define ofi_uring_sq_space_left(io_uring) 0
+#define ofi_uring_cq_ready(io_uring) 0
+#define ofi_uring_submit(io_uring) -FI_ENOSYS
+#define ofi_uring_peek_batch_cqe(io_uring, cqes, count) 0
+#define ofi_uring_cq_advance(io_uring, count) do {} while(0)
+#endif
 
 /*
  * Byte queue - streaming socket staging buffer
@@ -166,6 +465,19 @@ static inline size_t ofi_byteq_writeable(struct ofi_byteq *byteq)
 	return byteq->size - byteq->tail;
 }
 
+static inline void ofi_byteq_consume(struct ofi_byteq *byteq, size_t bytes)
+{
+	if (bytes == ofi_byteq_readable(byteq))
+		ofi_byteq_discard(byteq);
+	else
+		byteq->head += (unsigned) bytes;
+}
+
+static inline void ofi_byteq_add(struct ofi_byteq *byteq, size_t bytes)
+{
+	byteq->tail += (unsigned) bytes;
+}
+
 static inline size_t
 ofi_byteq_read(struct ofi_byteq *byteq, void *buf, size_t len)
 {
@@ -177,13 +489,12 @@ ofi_byteq_read(struct ofi_byteq *byteq, void *buf, size_t len)
 
 	if (len < avail) {
 		memcpy(buf, &byteq->data[byteq->head], len);
-		byteq->head += len;
+		byteq->head += (unsigned) len;
 		return len;
 	}
 
 	memcpy(buf, &byteq->data[byteq->head], avail);
-	byteq->head = 0;
-	byteq->tail = 0;
+	ofi_byteq_discard(byteq);
 	return avail;
 }
 
@@ -192,67 +503,46 @@ ofi_byteq_write(struct ofi_byteq *byteq, const void *buf, size_t len)
 {
 	assert(len <= ofi_byteq_writeable(byteq));
 	memcpy(&byteq->data[byteq->tail], buf, len);
-	byteq->tail += len;
+	ofi_byteq_add(byteq, len);
 }
 
 void ofi_byteq_writev(struct ofi_byteq *byteq, const struct iovec *iov,
 		      size_t cnt);
 
-static inline ssize_t ofi_byteq_recv(struct ofi_byteq *byteq, SOCKET sock)
-{
-	size_t avail;
-	ssize_t ret;
-
-	avail = ofi_byteq_writeable(byteq);
-	assert(avail);
-	ret = ofi_recv_socket(sock, &byteq->data[byteq->tail], avail,
-			      MSG_NOSIGNAL);
-	if (ret > 0)
-		byteq->tail += ret;
-	return ret;
-}
-
 size_t ofi_byteq_readv(struct ofi_byteq *byteq, struct iovec *iov,
 		       size_t cnt, size_t offset);
 
-static inline ssize_t ofi_byteq_send(struct ofi_byteq *byteq, SOCKET sock)
-{
-	size_t avail;
-	ssize_t ret;
-
-	avail = ofi_byteq_readable(byteq);
-	assert(avail);
-	ret = ofi_send_socket(sock, &byteq->data[byteq->head], avail,
-			      MSG_NOSIGNAL);
-	if (ret == avail) {
-		byteq->head = 0;
-		byteq->tail = 0;
-	} else if (ret > 0) {
-		byteq->head += ret;
-	}
-	return ret;
-}
-
 
 /*
  * Buffered socket - socket with send/receive staging buffers.
  */
 struct ofi_bsock {
 	SOCKET sock;
+	struct ofi_sockapi *sockapi;
+	struct ofi_sockctx tx_sockctx;
+	struct ofi_sockctx rx_sockctx;
+	struct ofi_sockctx cancel_sockctx;
 	struct ofi_byteq sq;
 	struct ofi_byteq rq;
 	size_t zerocopy_size;
 	uint32_t async_index;
 	uint32_t done_index;
+	bool async_prefetch;
 };
 
 static inline void
-ofi_bsock_init(struct ofi_bsock *bsock, ssize_t sbuf_size, ssize_t rbuf_size)
+ofi_bsock_init(struct ofi_bsock *bsock, struct ofi_sockapi *sockapi,
+	       ssize_t sbuf_size, ssize_t rbuf_size, void *context)
 {
 	bsock->sock = INVALID_SOCKET;
+	bsock->sockapi = sockapi;
+	ofi_sockctx_init(&bsock->tx_sockctx, context);
+	ofi_sockctx_init(&bsock->rx_sockctx, context);
+	ofi_sockctx_init(&bsock->cancel_sockctx, context);
 	ofi_byteq_init(&bsock->sq, sbuf_size);
 	ofi_byteq_init(&bsock->rq, rbuf_size);
 	bsock->zerocopy_size = SIZE_MAX;
+	bsock->async_prefetch = false;
 
 	/* first async op will wrap back to 0 as the starting index */
 	bsock->async_index = UINT32_MAX;
@@ -275,18 +565,36 @@ static inline size_t ofi_bsock_tosend(struct ofi_bsock *bsock)
 	return ofi_byteq_readable(&bsock->sq);
 }
 
-ssize_t ofi_bsock_flush(struct ofi_bsock *bsock);
-/* For sends started asynchronously, the return value will be -EINPROGRESS,
+static inline int ofi_bsock_connect(struct ofi_bsock *bsock,
+				    const struct sockaddr *addr,
+				    socklen_t addrlen)
+{
+	return bsock->sockapi->connect(bsock->sockapi, bsock->sock,
+				       addr, addrlen, &bsock->tx_sockctx);
+}
+
+static inline int ofi_bsock_recv_unbuffered(struct ofi_bsock *bsock, void *buf,
+					    size_t len)
+{
+	assert(!ofi_bsock_readable(bsock));
+	return bsock->sockapi->recv(bsock->sockapi, bsock->sock, buf, len,
+				    MSG_NOSIGNAL, &bsock->rx_sockctx);
+}
+
+int ofi_bsock_flush(struct ofi_bsock *bsock);
+int ofi_bsock_flush_sync(struct ofi_bsock *bsock);
+/* For sends started asynchronously, the return value will be -EINPROGRESS_ASYNC,
  * and len will be set to the number of bytes that were queued.
  */
-ssize_t ofi_bsock_send(struct ofi_bsock *bsock, const void *buf, size_t *len);
-ssize_t ofi_bsock_sendv(struct ofi_bsock *bsock, const struct iovec *iov,
-			size_t cnt, size_t *len);
-ssize_t ofi_bsock_recv(struct ofi_bsock *bsock, void *buf, size_t len);
-ssize_t ofi_bsock_recvv(struct ofi_bsock *bsock, struct iovec *iov,
-			size_t cnt);
+int ofi_bsock_send(struct ofi_bsock *bsock, const void *buf, size_t *len);
+int ofi_bsock_sendv(struct ofi_bsock *bsock, const struct iovec *iov,
+		    size_t cnt, size_t *len);
+int ofi_bsock_recv(struct ofi_bsock *bsock, void *buf, size_t *len);
+int ofi_bsock_recvv(struct ofi_bsock *bsock, struct iovec *iov,
+		    size_t cnt, size_t *len);
 uint32_t ofi_bsock_async_done(const struct fi_provider *prov,
 			      struct ofi_bsock *bsock);
+void ofi_bsock_prefetch_done(struct ofi_bsock *bsock, size_t len);
 
 
 /*
@@ -401,6 +709,23 @@ static inline int ofi_translate_addr_format(int family)
 	}
 }
 
+static inline size_t ofi_sizeof_addr_format(int format)
+{
+	switch (format) {
+	case FI_SOCKADDR:
+		return sizeof(union ofi_sock_ip);
+	case FI_SOCKADDR_IN:
+		return sizeof(struct sockaddr_in);
+	case FI_SOCKADDR_IN6:
+		return sizeof(struct sockaddr_in6);
+	case FI_SOCKADDR_IB:
+		return sizeof(struct ofi_sockaddr_ib);
+	default:
+		FI_WARN(&core_prov, FI_LOG_CORE, "Unsupported address format\n");
+		return 0;
+	}
+}
+
 uint16_t ofi_get_sa_family(const struct fi_info *info);
 
 static inline bool ofi_sin_is_any_addr(const struct sockaddr *sa)
@@ -510,7 +835,12 @@ static inline void * ofi_get_ipaddr(const struct sockaddr *addr)
 
 static inline bool ofi_valid_dest_ipaddr(const struct sockaddr *addr)
 {
-	return ofi_addr_get_port(addr) && !ofi_is_any_addr(addr);
+	char sin_zero[8] = {0};
+
+	return ofi_addr_get_port(addr) && !ofi_is_any_addr(addr) &&
+	       (addr->sa_family != AF_INET ||
+	         !memcmp(((const struct sockaddr_in *) addr)->sin_zero,
+			 sin_zero, sizeof sin_zero));
 }
 
 static inline bool ofi_equals_ipaddr(const struct sockaddr *addr1,
