@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2013-2021 Intel Corporation. All rights reserved
+ * (C) Copyright 2021 Amazon.com, Inc. or its affiliates.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -37,12 +38,12 @@
 #include "ofi_iov.h"
 #include "smr.h"
 
-
-static void smr_format_rma_iov(struct smr_cmd *cmd, const struct fi_rma_iov *rma_iov,
-			       size_t iov_count)
+static void smr_add_rma_cmd(struct smr_region *peer_smr,
+		const struct fi_rma_iov *rma_iov, size_t iov_count,
+		struct smr_cmd_entry *ce)
 {
-	cmd->rma.rma_count = iov_count;
-	memcpy(cmd->rma.rma_iov, rma_iov, sizeof(*rma_iov) * iov_count);
+	ce->rma_cmd.rma.rma_count = iov_count;
+	memcpy(ce->rma_cmd.rma.rma_iov, rma_iov, sizeof(*rma_iov) * iov_count);
 }
 
 static void smr_format_rma_resp(struct smr_cmd *cmd, fi_addr_t peer_id,
@@ -53,15 +54,22 @@ static void smr_format_rma_resp(struct smr_cmd *cmd, fi_addr_t peer_id,
 	cmd->msg.hdr.size = total_len;
 }
 
-ssize_t smr_rma_fast(struct smr_region *peer_smr, struct smr_cmd *cmd,
-		     const struct iovec *iov, size_t iov_count,
-		     const struct fi_rma_iov *rma_iov, size_t rma_count,
-		     void **desc, int peer_id, void *context, uint32_t op,
-		     uint64_t op_flags)
+static ssize_t smr_rma_fast(struct smr_region *peer_smr, const struct iovec *iov,
+			size_t iov_count, const struct fi_rma_iov *rma_iov,
+			size_t rma_count, void **desc, int peer_id, void *context,
+			uint32_t op, uint64_t op_flags)
 {
 	struct iovec cma_iovec[SMR_IOV_LIMIT], rma_iovec[SMR_IOV_LIMIT];
+	struct smr_cmd_entry *ce;
 	size_t total_len;
 	int ret, i;
+	int64_t pos;
+
+	ret = smr_cmd_queue_next(smr_cmd_queue(peer_smr), &ce, &pos);
+	if (ret == -FI_ENOENT) {
+		ret = -FI_EAGAIN;
+		goto signal;
+	}
 
 	memcpy(cma_iovec, iov, sizeof(*iov) * iov_count);
 	for (i = 0; i < rma_count; i++) {
@@ -75,37 +83,42 @@ ssize_t smr_rma_fast(struct smr_region *peer_smr, struct smr_cmd *cmd,
 			   rma_count, 0, total_len, op == ofi_op_write);
 
 	if (ret)
-		return ret;
+		goto discard_cmd;
 
-	smr_format_rma_resp(cmd, peer_id, rma_iov, rma_count, total_len,
+	smr_format_rma_resp(&ce->cmd, peer_id, rma_iov, rma_count, total_len,
 			    (op == ofi_op_write) ? ofi_op_write_async :
 			    ofi_op_read_async, op_flags);
+	smr_cmd_queue_commit(ce, pos);
 
 	return 0;
+
+discard_cmd:
+	smr_cmd_queue_discard(ce, pos);
+signal:
+	smr_signal(peer_smr);
+	return ret;
 }
 
-ssize_t smr_generic_rma(struct smr_ep *ep, const struct iovec *iov,
+static ssize_t smr_generic_rma(struct smr_ep *ep, const struct iovec *iov,
 	size_t iov_count, const struct fi_rma_iov *rma_iov, size_t rma_count,
 	void **desc, fi_addr_t addr, void *context, uint32_t op, uint64_t data,
 	uint64_t op_flags)
 {
 	struct smr_domain *domain;
 	struct smr_region *peer_smr;
-	struct smr_inject_buf *tx_buf;
-	struct smr_sar_msg *sar;
-	struct smr_resp *resp;
-	struct smr_cmd *cmd;
-	struct smr_tx_entry *pend;
-	enum fi_hmem_iface iface;
-	uint64_t device;
+	struct ofi_mr *smr_desc;
 	int64_t id, peer_id;
-	int cmds, err = 0, comp = 1;
-	uint16_t comp_flags;
+	int cmds, err = 0, proto = smr_src_inline;
 	ssize_t ret = 0;
 	size_t total_len;
+	bool use_ipc = false;
+	struct smr_cmd_entry *ce;
+	int64_t pos;
 
 	assert(iov_count <= SMR_IOV_LIMIT);
 	assert(rma_count <= SMR_IOV_LIMIT);
+	assert(ofi_total_iov_len(iov, iov_count) ==
+	       ofi_total_rma_iov_len(rma_iov, rma_count));
 
 	domain = container_of(ep->util_ep.domain, struct smr_domain, util_domain);
 
@@ -120,132 +133,86 @@ ssize_t smr_generic_rma(struct smr_ep *ep, const struct iovec *iov,
 		    (FI_REMOTE_CQ_DATA | FI_DELIVERY_COMPLETE)) &&
 		     rma_count == 1 && smr_cma_enabled(ep, peer_smr));
 
-	fastlock_acquire(&peer_smr->lock);
-	if (peer_smr->cmd_cnt < cmds ||
-	    smr_peer_data(ep->region)[id].sar_status) {
-		ret = -FI_EAGAIN;
-		goto unlock_region;
-	}
+	if (smr_peer_data(ep->region)[id].sar_status)
+		return -FI_EAGAIN;
 
-	fastlock_acquire(&ep->util_ep.tx_cq->cq_lock);
-	if (ofi_cirque_isfull(ep->util_ep.tx_cq->cirq)) {
-		ret = -FI_EAGAIN;
-		goto unlock_cq;
-	}
-
-	cmd = ofi_cirque_next(smr_cmd_queue(peer_smr));
+	ofi_spin_lock(&ep->tx_lock);
 
 	if (cmds == 1) {
-		err = smr_rma_fast(peer_smr, cmd, iov, iov_count, rma_iov,
+		err = smr_rma_fast(peer_smr, iov, iov_count, rma_iov,
 				   rma_count, desc, peer_id,  context, op,
 				   op_flags);
-		comp_flags = cmd->msg.hdr.op_flags;
-		goto commit_comp;
+		if (err) {
+			FI_WARN(&smr_prov, FI_LOG_EP_CTRL,
+				"error doing fast RMA\n");
+			if (err == -FI_EAGAIN) {
+				ret = -FI_EAGAIN;
+				goto signal;
+			}
+
+			ret = smr_write_err_comp(ep->util_ep.rx_cq, NULL,
+						op_flags, 0, -err);
+		} else {
+			ret = smr_complete_tx(ep, context, op, op_flags);
+		}
+
+		if (ret) {
+			FI_WARN(&smr_prov, FI_LOG_EP_CTRL,
+				"unable to process tx completion\n");
+		}
+		goto signal;
 	}
 
-	iface = smr_get_mr_hmem_iface(ep->util_ep.domain, desc, &device);
+	ret = smr_cmd_queue_next(smr_cmd_queue(peer_smr), &ce, &pos);
+	if (ret == -FI_ENOENT) {
+		/* kick the peer to process any outstanding commands */
+		ret = -FI_EAGAIN;
+		goto signal;
+	}
 
 	total_len = ofi_total_iov_len(iov, iov_count);
+	assert(!(op_flags & FI_INJECT) || total_len <= SMR_INJECT_SIZE);
 
-	smr_generic_format(cmd, peer_id, op, 0, data, op_flags);
-	if (total_len <= SMR_MSG_DATA_LEN && op == ofi_op_write &&
-	    !(op_flags & FI_DELIVERY_COMPLETE)) {
-		smr_format_inline(cmd, iface, device, iov, iov_count);
-	} else if (total_len <= SMR_INJECT_SIZE &&
-		   !(op_flags & FI_DELIVERY_COMPLETE)) {
-		tx_buf = smr_freestack_pop(smr_inject_pool(peer_smr));
-		smr_format_inject(cmd, iface, device, iov, iov_count, peer_smr, tx_buf);
-		if (op == ofi_op_read_req) {
-			if (ofi_cirque_isfull(smr_resp_queue(ep->region))) {
-				smr_freestack_push(smr_inject_pool(peer_smr), tx_buf);
-				ret = -FI_EAGAIN;
-				goto unlock_cq;
-			}
-			cmd->msg.hdr.op_flags |= SMR_RMA_REQ;
-			resp = ofi_cirque_next(smr_resp_queue(ep->region));
-			pend = ofi_freestack_pop(ep->pend_fs);
-			smr_format_pend_resp(pend, cmd, context, iface, device, iov,
-					     iov_count, id, resp);
-			cmd->msg.hdr.data = smr_get_offset(ep->region, resp);
-			ofi_cirque_commit(smr_resp_queue(ep->region));
-			comp = 0;
-		}
-	} else {
-		if (ofi_cirque_isfull(smr_resp_queue(ep->region))) {
-			ret = -FI_EAGAIN;
-			goto unlock_cq;
-		}
-		resp = ofi_cirque_next(smr_resp_queue(ep->region));
-		pend = ofi_freestack_pop(ep->pend_fs);
-		if (smr_cma_enabled(ep, peer_smr) && iface == FI_HMEM_SYSTEM) {
-			smr_format_iov(cmd, iov, iov_count, total_len, ep->region,
-				       resp);
-		} else {
-			if (iface == FI_HMEM_ZE &&
-			    (smr_get_mr_flags(desc) & FI_HMEM_DEVICE_ONLY) &&
-			    iov_count == 1 &&
-			    smr_ze_ipc_enabled(ep->region, peer_smr)) {
-				ret = smr_format_ze_ipc(ep, id, cmd, iov,
-					device, total_len, ep->region,
-					resp, pend);
-			} else if (total_len <= smr_env.sar_threshold ||
-			    iface != FI_HMEM_SYSTEM) {
-				if (!peer_smr->sar_cnt) {
-					ret = -FI_EAGAIN;
-				} else {
-					sar = smr_freestack_pop(smr_sar_pool(peer_smr));
-					smr_format_sar(cmd, iface, device, iov,
-						       iov_count, total_len,
-						       ep->region, peer_smr, sar,
-						       pend, resp);
-					peer_smr->sar_cnt--;
-					smr_peer_data(ep->region)[id].sar_status = 1;
-				}
-			} else {
-				ret = smr_format_mmap(ep, cmd, iov, iov_count,
-						      total_len, pend, resp);
-			}
-			if (ret) {
-				ofi_freestack_push(ep->pend_fs, pend);
-				ret = -FI_EAGAIN;
-				goto unlock_cq;
-			}
-		}
-		smr_format_pend_resp(pend, cmd, context, iface, device, iov,
-				     iov_count, id, resp);
-		ofi_cirque_commit(smr_resp_queue(ep->region));
-		comp = 0;
+	/* Do not inline/inject if IPC is available so device to device
+	 * transfer may occur if possible. */
+	if (iov_count == 1 && desc && desc[0]) {
+		smr_desc = (struct ofi_mr *) *desc;
+		use_ipc = ofi_hmem_is_ipc_enabled(((struct ofi_mr *) *desc)->iface) &&
+				smr_desc->flags & FI_HMEM_DEVICE_ONLY &&
+				!(op_flags & FI_INJECT);
+	}
+	proto = smr_select_proto(use_ipc, smr_cma_enabled(ep, peer_smr), op,
+				 total_len, op_flags);
+
+	ret = smr_proto_ops[proto](ep, peer_smr, id, peer_id, op, 0, data,
+				   op_flags, (struct ofi_mr **)desc, iov,
+				   iov_count, total_len, context, &ce->cmd);
+	if (ret) {
+		smr_cmd_queue_discard(ce, pos);
+		goto signal;
 	}
 
-	comp_flags = cmd->msg.hdr.op_flags;
-	ofi_cirque_commit(smr_cmd_queue(peer_smr));
-	peer_smr->cmd_cnt--;
-	cmd = ofi_cirque_next(smr_cmd_queue(peer_smr));
-	smr_format_rma_iov(cmd, rma_iov, rma_count);
+	smr_add_rma_cmd(peer_smr, rma_iov, rma_count, ce);
+	smr_cmd_queue_commit(ce, pos);
 
-commit_comp:
-	ofi_cirque_commit(smr_cmd_queue(peer_smr));
-	peer_smr->cmd_cnt--;
-	smr_signal(peer_smr);
+	if (proto != smr_src_inline && proto != smr_src_inject)
+		goto signal;
 
-	if (!comp)
-		goto unlock_cq;
-
-	ret = smr_complete_tx(ep, context, op, comp_flags, err);
+	ret = smr_complete_tx(ep, context, op, op_flags);
 	if (ret) {
 		FI_WARN(&smr_prov, FI_LOG_EP_CTRL,
 			"unable to process tx completion\n");
 	}
 
-unlock_cq:
-	fastlock_release(&ep->util_ep.tx_cq->cq_lock);
-unlock_region:
-	fastlock_release(&peer_smr->lock);
+signal:
+	smr_signal(peer_smr);
+	ofi_spin_unlock(&ep->tx_lock);
 	return ret;
 }
 
-ssize_t smr_read(struct fid_ep *ep_fid, void *buf, size_t len, void *desc,
-	fi_addr_t src_addr, uint64_t addr, uint64_t key, void *context)
+static ssize_t smr_read(struct fid_ep *ep_fid, void *buf, size_t len,
+			void *desc, fi_addr_t src_addr, uint64_t addr,
+			uint64_t key, void *context)
 {
 	struct smr_ep *ep;
 	struct iovec msg_iov;
@@ -264,9 +231,9 @@ ssize_t smr_read(struct fid_ep *ep_fid, void *buf, size_t len, void *desc,
 			       smr_ep_tx_flags(ep));
 }
 
-ssize_t smr_readv(struct fid_ep *ep_fid, const struct iovec *iov, void **desc,
-	size_t count, fi_addr_t src_addr, uint64_t addr, uint64_t key,
-	void *context)
+static ssize_t smr_readv(struct fid_ep *ep_fid, const struct iovec *iov,
+			 void **desc, size_t count, fi_addr_t src_addr,
+			 uint64_t addr, uint64_t key, void *context)
 {
 	struct smr_ep *ep;
 	struct fi_rma_iov rma_iov;
@@ -282,8 +249,8 @@ ssize_t smr_readv(struct fid_ep *ep_fid, const struct iovec *iov, void **desc,
 			       smr_ep_tx_flags(ep));
 }
 
-ssize_t smr_readmsg(struct fid_ep *ep_fid, const struct fi_msg_rma *msg,
-	uint64_t flags)
+static ssize_t smr_readmsg(struct fid_ep *ep_fid, const struct fi_msg_rma *msg,
+			   uint64_t flags)
 {
 	struct smr_ep *ep;
 
@@ -296,8 +263,9 @@ ssize_t smr_readmsg(struct fid_ep *ep_fid, const struct fi_msg_rma *msg,
 			       flags | ep->util_ep.tx_msg_flags);
 }
 
-ssize_t smr_write(struct fid_ep *ep_fid, const void *buf, size_t len, void *desc,
-	fi_addr_t dest_addr, uint64_t addr, uint64_t key, void *context)
+static ssize_t smr_write(struct fid_ep *ep_fid, const void *buf, size_t len,
+			 void *desc, fi_addr_t dest_addr, uint64_t addr,
+			 uint64_t key, void *context)
 {
 	struct smr_ep *ep;
 	struct iovec msg_iov;
@@ -316,9 +284,9 @@ ssize_t smr_write(struct fid_ep *ep_fid, const void *buf, size_t len, void *desc
 			       smr_ep_tx_flags(ep));
 }
 
-ssize_t smr_writev(struct fid_ep *ep_fid, const struct iovec *iov, void **desc,
-		size_t count, fi_addr_t dest_addr, uint64_t addr, uint64_t key,
-		void *context)
+static ssize_t smr_writev(struct fid_ep *ep_fid, const struct iovec *iov,
+			  void **desc, size_t count, fi_addr_t dest_addr,
+			  uint64_t addr, uint64_t key, void *context)
 {
 	struct smr_ep *ep;
 	struct fi_rma_iov rma_iov;
@@ -335,8 +303,8 @@ ssize_t smr_writev(struct fid_ep *ep_fid, const struct iovec *iov, void **desc,
 }
 
 
-ssize_t smr_writemsg(struct fid_ep *ep_fid, const struct fi_msg_rma *msg,
-	uint64_t flags)
+static ssize_t smr_writemsg(struct fid_ep *ep_fid, const struct fi_msg_rma *msg,
+			    uint64_t flags)
 {
 	struct smr_ep *ep;
 
@@ -349,20 +317,20 @@ ssize_t smr_writemsg(struct fid_ep *ep_fid, const struct fi_msg_rma *msg,
 			       flags | ep->util_ep.tx_msg_flags);
 }
 
-ssize_t smr_generic_rma_inject(struct fid_ep *ep_fid, const void *buf,
-	size_t len, fi_addr_t dest_addr, uint64_t addr, uint64_t key,
-	uint64_t data, uint64_t flags)
+static ssize_t smr_generic_rma_inject(struct fid_ep *ep_fid, const void *buf,
+		size_t len, fi_addr_t dest_addr, uint64_t addr, uint64_t key,
+		uint64_t data, uint64_t flags)
 {
 	struct smr_ep *ep;
 	struct smr_domain *domain;
 	struct smr_region *peer_smr;
-	struct smr_inject_buf *tx_buf;
-	struct smr_cmd *cmd;
 	struct iovec iov;
 	struct fi_rma_iov rma_iov;
 	int64_t id, peer_id;
-	int cmds;
+	int cmds, proto = smr_src_inline;
 	ssize_t ret = 0;
+	struct smr_cmd_entry *ce;
+	int64_t pos;
 
 	assert(len <= SMR_INJECT_SIZE);
 	ep = container_of(ep_fid, struct smr_ep, util_ep.ep_fid.fid);
@@ -378,12 +346,8 @@ ssize_t smr_generic_rma_inject(struct fid_ep *ep_fid, const void *buf,
 	cmds = 1 + !(domain->fast_rma && !(flags & FI_REMOTE_CQ_DATA) &&
 		     smr_cma_enabled(ep, peer_smr));
 
-	fastlock_acquire(&peer_smr->lock);
-	if (peer_smr->cmd_cnt < cmds ||
-	    smr_peer_data(ep->region)[id].sar_status) {
-		ret = -FI_EAGAIN;
-		goto unlock_region;
-	}
+	if (smr_peer_data(ep->region)[id].sar_status)
+		return -FI_EAGAIN;
 
 	iov.iov_base = (void *) buf;
 	iov.iov_len = len;
@@ -391,43 +355,37 @@ ssize_t smr_generic_rma_inject(struct fid_ep *ep_fid, const void *buf,
 	rma_iov.len = len;
 	rma_iov.key = key;
 
-	cmd = ofi_cirque_next(smr_cmd_queue(peer_smr));
-
 	if (cmds == 1) {
-		ret = smr_rma_fast(peer_smr, cmd, &iov, 1, &rma_iov, 1, NULL,
+		ret = smr_rma_fast(peer_smr, &iov, 1, &rma_iov, 1, NULL,
 				   peer_id, NULL, ofi_op_write, flags);
 		if (ret)
-			goto unlock_region;
-		goto commit;
+			return ret;
+		goto signal;
 	}
 
-	smr_generic_format(cmd, peer_id, ofi_op_write, 0, data, flags);
-	if (len <= SMR_MSG_DATA_LEN) {
-		smr_format_inline(cmd, FI_HMEM_SYSTEM, 0, &iov, 1);
-	} else {
-		tx_buf = smr_freestack_pop(smr_inject_pool(peer_smr));
-		smr_format_inject(cmd, FI_HMEM_SYSTEM, 0, &iov, 1,
-				  peer_smr, tx_buf);
+	ret = smr_cmd_queue_next(smr_cmd_queue(peer_smr), &ce, &pos);
+	if (ret == -FI_ENOENT) {
+		/* kick the peer to process any outstanding commands */
+		smr_signal(peer_smr);
+		return -FI_EAGAIN;
 	}
 
-	ofi_cirque_commit(smr_cmd_queue(peer_smr));
-	peer_smr->cmd_cnt--;
-	cmd = ofi_cirque_next(smr_cmd_queue(peer_smr));
-	smr_format_rma_iov(cmd, &rma_iov, 1);
+	proto = len <= SMR_MSG_DATA_LEN ? smr_src_inline : smr_src_inject;
+	ret = smr_proto_ops[proto](ep, peer_smr, id, peer_id, ofi_op_write, 0,
+			data, flags, NULL, &iov, 1, len, NULL, &ce->cmd);
 
-commit:
-	ofi_cirque_commit(smr_cmd_queue(peer_smr));
-	peer_smr->cmd_cnt--;
+	assert(!ret);
+	smr_add_rma_cmd(peer_smr, &rma_iov, 1, ce);
+	smr_cmd_queue_commit(ce, pos);
+signal:
 	smr_signal(peer_smr);
 	ofi_ep_tx_cntr_inc_func(&ep->util_ep, ofi_op_write);
-unlock_region:
-	fastlock_release(&peer_smr->lock);
 	return ret;
 }
 
-ssize_t smr_writedata(struct fid_ep *ep_fid, const void *buf, size_t len,
-		      void *desc, uint64_t data, fi_addr_t dest_addr,
-		      uint64_t addr, uint64_t key, void *context)
+static ssize_t smr_writedata(struct fid_ep *ep_fid, const void *buf, size_t len,
+			     void *desc, uint64_t data, fi_addr_t dest_addr,
+			     uint64_t addr, uint64_t key, void *context)
 {
 	struct smr_ep *ep;
 	struct iovec iov;
@@ -446,16 +404,17 @@ ssize_t smr_writedata(struct fid_ep *ep_fid, const void *buf, size_t len,
 			       FI_REMOTE_CQ_DATA | smr_ep_tx_flags(ep));
 }
 
-ssize_t smr_rma_inject(struct fid_ep *ep_fid, const void *buf,
-	size_t len, fi_addr_t dest_addr, uint64_t addr, uint64_t key)
+static ssize_t smr_rma_inject(struct fid_ep *ep_fid, const void *buf,
+			      size_t len, fi_addr_t dest_addr, uint64_t addr,
+			      uint64_t key)
 {
 	return smr_generic_rma_inject(ep_fid, buf, len, dest_addr, addr, key,
 				      0, 0);
 }
 
-ssize_t smr_inject_writedata(struct fid_ep *ep_fid, const void *buf, size_t len,
-			     uint64_t data, fi_addr_t dest_addr, uint64_t addr,
-			     uint64_t key)
+static ssize_t smr_inject_writedata(struct fid_ep *ep_fid, const void *buf,
+			size_t len, uint64_t data, fi_addr_t dest_addr,
+			uint64_t addr, uint64_t key)
 {
 	return smr_generic_rma_inject(ep_fid, buf, len, dest_addr, addr, key,
 				      data, FI_REMOTE_CQ_DATA);
