@@ -1,5 +1,5 @@
 /**
-* Copyright (C) NVIDIA 2021.  ALL RIGHTS RESERVED.
+* Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2021. ALL RIGHTS RESERVED.
 *
 * See file LICENSE for terms.
 */
@@ -47,9 +47,6 @@ static void usage(const struct perftest_context *ctx, const char *program)
 #if defined (HAVE_MPI)
     printf("  Note: test can be also launched as an MPI application\n");
     printf("\n");
-#elif defined (HAVE_RTE)
-    printf("  Note: this test can be also launched as an libRTE application\n");
-    printf("\n");
 #endif
     printf("  Usage: %s [ server-hostname ] [ options ]\n", program);
     printf("\n");
@@ -137,6 +134,8 @@ static void usage(const struct perftest_context *ctx, const char *program)
     printf("                        sleep      : go to sleep after posting requests\n");
     printf("     -H <size>      active message header size (%zu), not included in message size\n",
                                 ctx->params.super.ucp.am_hdr_size);
+    printf("     -y             do additional memcopy to the user memory in active message receive handler\n");
+    printf("     -z             pass pre-registered memory handle\n");
     printf("\n");
     printf("   NOTE: When running UCP tests, transport and device should be specified by\n");
     printf("         environment variables: UCX_TLS and UCX_[SELF|SHM|NET]_DEVICES.\n");
@@ -416,6 +415,12 @@ ucs_status_t parse_test_params(perftest_params_t *params, char opt,
             return UCS_ERR_INVALID_PARAM;
         }
         return UCS_OK;
+    case 'y':
+        params->super.flags |= UCX_PERF_TEST_FLAG_AM_RECV_COPY;
+        return UCS_OK;
+    case 'z':
+        params->super.flags |= UCX_PERF_TEST_FLAG_PREREG;
+        return UCS_OK;
     default:
        return UCS_ERR_INVALID_PARAM;
     }
