@@ -1,5 +1,5 @@
 /**
- * Copyright (C) Mellanox Technologies Ltd. 2020.  ALL RIGHTS RESERVED.
+ * Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2020. ALL RIGHTS RESERVED.
  *
  * See file LICENSE for terms.
  */
@@ -19,7 +19,8 @@
     })
 
 
-ucs_status_t ucp_tag_send_start_rndv(ucp_request_t *req);
+ucs_status_t
+ucp_tag_send_start_rndv(ucp_request_t *req, const ucp_request_param_t *param);
 
 void ucp_tag_rndv_matched(ucp_worker_h worker, ucp_request_t *req,
                           const ucp_rndv_rts_hdr_t *rts_hdr, size_t hdr_length);
@@ -30,6 +31,8 @@ ucs_status_t ucp_tag_rndv_process_rts(ucp_worker_h worker,
 
 size_t ucp_tag_rndv_rts_pack(void *dest, void *arg);
 
+ucs_status_t ucp_proto_progress_tag_rndv_rts(uct_pending_req_t *self);
+
 
 static UCS_F_ALWAYS_INLINE ucp_rndv_rts_hdr_t *
 ucp_tag_rndv_rts_from_rdesc(ucp_recv_desc_t *rdesc)
