@@ -1,5 +1,5 @@
 /**
- * Copyright (C) Mellanox Technologies Ltd. 2018.  ALL RIGHTS RESERVED.
+ * Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2018. ALL RIGHTS RESERVED.
  *
  * See file LICENSE for terms.
  */
@@ -13,7 +13,7 @@
 #include <ucs/datastruct/list.h>
 #include <ucs/stats/stats_fwd.h>
 #include <ucs/sys/compiler_def.h>
-#include <ucs/sys/topo.h>
+#include <ucs/sys/topo/base/topo.h>
 #include <pthread.h>
 
 
@@ -29,27 +29,23 @@ extern ucs_memtype_cache_t *ucs_memtype_cache_global_instance;
 
 /* Memory information record */
 typedef struct ucs_memory_info {
-    ucs_memory_type_t type;          /**< Memory type, use uint8 for compact size */
+    ucs_memory_type_t type;          /**< Memory type */
     ucs_sys_device_t  sys_dev;       /**< System device index */
     void              *base_address; /**< Base address of the underlying allocation */
     size_t            alloc_length;  /**< Whole length of the underlying allocation */
 } ucs_memory_info_t;
 
 
-struct ucs_memtype_cache_region {
-    ucs_pgt_region_t    super;    /**< Base class - page table region */
-    ucs_list_link_t     list;     /**< List element */
-    ucs_memory_info_t   mem_info; /**< Memory type and system device the address
-                                       belongs to */
-};
-
-
 struct ucs_memtype_cache {
     pthread_rwlock_t      lock;       /**< protests the page table */
     ucs_pgtable_t         pgtable;    /**< Page table to hold the regions */
 };
 
 
+void ucs_memtype_cache_global_init();
+void ucs_memtype_cache_cleanup();
+
+
 /**
  * Find if address range is in memtype cache.
  *
@@ -60,7 +56,9 @@ struct ucs_memtype_cache {
  *                              means the memory type is an unknown non-host
  *                              memory, and should be detected in another way.
  *
- * @return Error code.
+ * @return UCS_OK              - an element was found and the memory info is valid.
+ * @return UCS_ERR_NO_ELEM     - an element was not found.
+ * @return UCS_ERR_UNSUPPORTED - the memory type cache is disabled.
  */
 ucs_status_t ucs_memtype_cache_lookup(const void *address, size_t size,
                                       ucs_memory_info_t *mem_info);
@@ -73,11 +71,14 @@ ucs_status_t ucs_memtype_cache_lookup(const void *address, size_t size,
  *
  * @param [in]  address         Start address to update.
  * @param [in]  size            Size of the memory to update.
- * @param [in]  mem_info        Set the memory info of the address range to this
+ * @param [in]  mem_type        Set the memory type of the address range to this
  *                              value.
+ * @param [in]  sys_dev         Set the system device of the address range to
+ *                              this value.
  */
 void ucs_memtype_cache_update(const void *address, size_t size,
-                              const ucs_memory_info_t *mem_info);
+                              ucs_memory_type_t mem_type,
+                              ucs_sys_device_t sys_dev);
 
 
 /**
@@ -89,14 +90,6 @@ void ucs_memtype_cache_update(const void *address, size_t size,
 void ucs_memtype_cache_remove(const void *address, size_t size);
 
 
-/**
- * Helper function to set memory info structure to host memory type.
- *
- * @param [out] mem_info        Pointer to memory info structure.
- */
-void ucs_memory_info_set_host(ucs_memory_info_t *mem_info);
-
-
 /**
  * Find if global memtype_cache is empty.
  *
@@ -108,6 +101,21 @@ static UCS_F_ALWAYS_INLINE int ucs_memtype_cache_is_empty()
            (ucs_memtype_cache_global_instance->pgtable.num_regions == 0);
 }
 
+
+/**
+ * Helper function to set memory info structure to host memory type.
+ *
+ * @param [out] mem_info        Pointer to memory info structure.
+ */
+static UCS_F_ALWAYS_INLINE void
+ucs_memory_info_set_host(ucs_memory_info_t *mem_info)
+{
+    mem_info->type         = UCS_MEMORY_TYPE_HOST;
+    mem_info->sys_dev      = UCS_SYS_DEVICE_ID_UNKNOWN;
+    mem_info->base_address = NULL;
+    mem_info->alloc_length = -1;
+}
+
 END_C_DECLS
 
 #endif
