@@ -1,5 +1,5 @@
 /**
-* Copyright (C) Mellanox Technologies Ltd. 2001-2017.  ALL RIGHTS RESERVED.
+* Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2001-2017. ALL RIGHTS RESERVED.
 * Copyright (C) UT-Battelle, LLC. 2015. ALL RIGHTS RESERVED.
 * Copyright (C) Arm, Ltd. 2021. ALL RIGHTS RESERVED.
 *
@@ -49,8 +49,10 @@
 #define UCS_F_CTOR __attribute__((constructor))
 #define UCS_F_DTOR __attribute__((destructor))
 
-/* Silence "defined but not used" error for static function */
-#define UCS_F_MAYBE_UNUSED __attribute__((used))
+/* Silence "defined but not used" error for static function,
+ * remove it by linker if it's not used at all.
+ */
+#define UCS_F_MAYBE_UNUSED __attribute__((unused))
 
 /* Non-null return */
 #define UCS_F_NON_NULL __attribute__((nonnull))
@@ -66,7 +68,7 @@
  * which can't optimize properly.
  */
 #if (((__GNUC__ == 4) && (__GNUC_MINOR__ == 1)) || !defined(__OPTIMIZE__))
-#  define UCS_V_INITIALIZED(_v)  (_v = (typeof(_v))0)
+#  define UCS_V_INITIALIZED(_v)  (_v = (ucs_typeof(_v))0)
 #else
 #  define UCS_V_INITIALIZED(_v)  ((void)0)
 #endif
@@ -100,7 +102,7 @@
 
 /* Helper macro to calculate an address with offset equal to size of _type */
 #define UCS_PTR_TYPE_OFFSET(_ptr, _type) \
-    ((void *)((typeof(_type) *)(_ptr) + 1))
+    ((void *)((ucs_typeof(_type) *)(_ptr) + 1))
 
 /* Helper macro to calculate ptr difference (_end - _start) */
 #define UCS_PTR_BYTE_DIFF(_start, _end) \
@@ -134,6 +136,17 @@
     ( (_type*)( (char*)(void*)(_ptr) - ucs_offsetof(_type, _member) )  )
 
 
+/**
+ * Get the type of a structure or variable.
+ * 
+ * @param _type  Return the type of this argument.
+ * 
+ * @return The type of the given argument.
+ */
+#define ucs_typeof(_type) \
+    __typeof__(_type)
+
+
 /**
  * @return Address of a derived structure. It must have a "super" member at offset 0.
  * NOTE: we use the built-in offsetof here because we can't use ucs_offsetof() in
@@ -161,7 +174,7 @@
  * @return Type of _field in _type.
  */
 #define ucs_field_type(_type, _field) \
-    typeof(((_type*)0)->_field)
+    ucs_typeof(((_type*)0)->_field)
 
 /**
  * Prevent compiler from reordering instructions
@@ -192,4 +205,29 @@
 #define UCS_STATIC_CLEANUP \
     static void UCS_F_DTOR UCS_PP_APPEND_UNIQUE_ID(ucs_initializer_dtor)()
 
+/*
+ * Check if the two types are the same
+ */
+#define ucs_same_type(_type1, _type2) \
+    __builtin_types_compatible_p(_type1, _type2)
+
+/*
+ * Iterate over all elements of a C-array
+ */
+#define ucs_carray_for_each(_elem, _array, _length) \
+    for ((_elem) = (_array); (_elem) < ((_array) + (_length)); ++(_elem))
+
+/*
+ * Swap two variables values
+ */
+#define ucs_swap(_a, _b) \
+    { \
+        ucs_typeof(*(_a)) __tmp; \
+        \
+        UCS_STATIC_ASSERT(ucs_same_type(ucs_typeof(*(_a)), ucs_typeof(*(_b)))); \
+        __tmp = *(_a); \
+        *(_a) = *(_b); \
+        *(_b) = __tmp; \
+    }
+
 #endif /* UCS_COMPILER_DEF_H */
