@@ -1,5 +1,5 @@
 /**
-* Copyright (C) Mellanox Technologies Ltd. 2001-2014.  ALL RIGHTS RESERVED.
+* Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2001-2014. ALL RIGHTS RESERVED.
 *
 * See file LICENSE for terms.
 */
@@ -12,14 +12,18 @@
 #include <ucs/sys/module.h>
 #include <ucs/arch/cpu.h>
 #include <ucs/config/parser.h>
+#include <ucs/config/ucm_opts.h>
 #include <ucs/debug/debug_int.h>
 #include <ucs/debug/log.h>
-#include <ucs/debug/memtrack.h>
+#include <ucs/debug/memtrack_int.h>
 #include <ucs/profile/profile.h>
+#include <ucs/memory/memtype_cache.h>
+#include <ucs/memory/numa.h>
 #include <ucs/stats/stats.h>
 #include <ucs/async/async.h>
+#include <ucs/sys/lib.h>
 #include <ucs/sys/sys.h>
-#include <ucs/sys/topo.h>
+#include <ucs/sys/topo/base/topo.h>
 #include <ucs/sys/math.h>
 
 
@@ -85,13 +89,15 @@ static void ucs_modules_load()
     UCS_MODULE_FRAMEWORK_LOAD(ucs, UCS_MODULE_LOAD_FLAG_GLOBAL);
 }
 
-static void UCS_F_CTOR ucs_init()
+void UCS_F_CTOR ucs_init()
 {
     ucs_status_t status;
 
     ucs_check_cpu_flags();
     ucs_log_early_init(); /* Must be called before all others */
     ucs_global_opts_init();
+    ucs_init_ucm_opts();
+    ucs_memtype_cache_global_init();
     ucs_cpu_init();
     ucs_log_init();
 #ifdef ENABLE_STATS
@@ -108,10 +114,11 @@ static void UCS_F_CTOR ucs_init()
     }
 
     ucs_async_global_init();
+    ucs_numa_init();
     ucs_topo_init();
     ucs_rand_seed_init();
-    ucs_debug("%s loaded at 0x%lx", ucs_debug_get_lib_path(),
-              ucs_debug_get_lib_base_addr());
+    ucs_debug("%s loaded at 0x%lx", ucs_sys_get_lib_path(),
+              ucs_sys_get_lib_base_addr());
     ucs_debug("cmd line: %s", ucs_get_process_cmdline());
     ucs_modules_load();
 }
@@ -119,6 +126,7 @@ static void UCS_F_CTOR ucs_init()
 static void UCS_F_DTOR ucs_cleanup(void)
 {
     ucs_topo_cleanup();
+    ucs_numa_cleanup();
     ucs_async_global_cleanup();
     ucs_profile_cleanup(ucs_profile_default_ctx);
     ucs_debug_cleanup(0);
@@ -126,5 +134,8 @@ static void UCS_F_DTOR ucs_cleanup(void)
 #ifdef ENABLE_STATS
     ucs_stats_cleanup();
 #endif
+    ucs_memtype_cache_cleanup();
+    ucs_cleanup_ucm_opts();
+    ucs_global_opts_cleanup();
     ucs_log_cleanup();
 }
