@@ -1,5 +1,5 @@
 /**
- * Copyright (C) Mellanox Technologies Ltd. 2017-2019.  ALL RIGHTS RESERVED.
+ * Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2017-2019. ALL RIGHTS RESERVED.
  * See file LICENSE for terms.
  */
 
@@ -12,12 +12,17 @@
 #include "cuda_copy_ep.h"
 
 #include <uct/cuda/base/cuda_iface.h>
+#include <uct/cuda/base/cuda_md.h>
 #include <ucs/type/class.h>
 #include <ucs/sys/string.h>
-#include <ucs/async/async.h>
+#include <ucs/async/eventfd.h>
 #include <ucs/arch/cpu.h>
 
 
+#define UCT_CUDA_COPY_IFACE_OVERHEAD 0
+#define UCT_CUDA_COPY_IFACE_LATENCY  ucs_linear_func_make(8e-6, 0)
+
+
 static ucs_config_field_t uct_cuda_copy_iface_config_table[] = {
 
     {"", "", NULL,
@@ -32,6 +37,10 @@ static ucs_config_field_t uct_cuda_copy_iface_config_table[] = {
      "Max number of cuda events. -1 is infinite",
      ucs_offsetof(uct_cuda_copy_iface_config_t, max_cuda_events), UCS_CONFIG_TYPE_UINT},
 
+    {"BW", "10000MBs",
+     "Effective memory bandwidth",
+     ucs_offsetof(uct_cuda_copy_iface_config_t, bandwidth), UCS_CONFIG_TYPE_BW},
+
     {NULL}
 };
 
@@ -64,7 +73,7 @@ static ucs_status_t uct_cuda_copy_iface_query(uct_iface_h tl_iface,
 {
     uct_cuda_copy_iface_t *iface = ucs_derived_of(tl_iface, uct_cuda_copy_iface_t);
 
-    uct_base_iface_query(&iface->super, iface_attr);
+    uct_base_iface_query(&iface->super.super, iface_attr);
 
     iface_attr->iface_addr_len          = sizeof(uct_cuda_copy_iface_addr_t);
     iface_attr->device_addr_len         = 0;
@@ -78,7 +87,7 @@ static ucs_status_t uct_cuda_copy_iface_query(uct_iface_h tl_iface,
 
     iface_attr->cap.event_flags         = UCT_IFACE_FLAG_EVENT_SEND_COMP |
                                           UCT_IFACE_FLAG_EVENT_RECV      |
-                                          UCT_IFACE_FLAG_EVENT_ASYNC_CB;
+                                          UCT_IFACE_FLAG_EVENT_FD;
 
     iface_attr->cap.put.max_short       = UINT_MAX;
     iface_attr->cap.put.max_bcopy       = 0;
@@ -105,9 +114,9 @@ static ucs_status_t uct_cuda_copy_iface_query(uct_iface_h tl_iface,
     iface_attr->cap.am.max_hdr          = 0;
     iface_attr->cap.am.max_iov          = 1;
 
-    iface_attr->latency                 = ucs_linear_func_make(8e-6, 0);
+    iface_attr->latency                 = UCT_CUDA_COPY_IFACE_LATENCY;
     iface_attr->bandwidth.dedicated     = 0;
-    iface_attr->bandwidth.shared        = UCT_CUDA_COPY_IFACE_DEFAULT_BANDWIDTH;
+    iface_attr->bandwidth.shared        = iface->config.bandwidth;
     iface_attr->overhead                = UCT_CUDA_COPY_IFACE_OVERHEAD;
     iface_attr->priority                = 0;
 
@@ -118,19 +127,24 @@ static ucs_status_t uct_cuda_copy_iface_flush(uct_iface_h tl_iface, unsigned fla
                                               uct_completion_t *comp)
 {
     uct_cuda_copy_iface_t *iface = ucs_derived_of(tl_iface, uct_cuda_copy_iface_t);
+    uct_cuda_copy_queue_desc_t *q_desc;
+    ucs_queue_iter_t iter;
 
     if (comp != NULL) {
         return UCS_ERR_UNSUPPORTED;
     }
 
-    if (ucs_queue_is_empty(&iface->outstanding_event_q[UCT_CUDA_COPY_STREAM_H2D]) &&
-        ucs_queue_is_empty(&iface->outstanding_event_q[UCT_CUDA_COPY_STREAM_D2H])) {
-        UCT_TL_IFACE_STAT_FLUSH(ucs_derived_of(tl_iface, uct_base_iface_t));
-        return UCS_OK;
+    ucs_queue_for_each_safe(q_desc, iter, &iface->active_queue, queue) {
+        if (!ucs_queue_is_empty(&q_desc->event_queue)) {
+            UCT_TL_IFACE_STAT_FLUSH_WAIT(ucs_derived_of(tl_iface,
+                                                        uct_base_iface_t));
+            return UCS_INPROGRESS;
+        }
     }
 
-    UCT_TL_IFACE_STAT_FLUSH_WAIT(ucs_derived_of(tl_iface, uct_base_iface_t));
-    return UCS_INPROGRESS;
+    UCT_TL_IFACE_STAT_FLUSH(ucs_derived_of(tl_iface, uct_base_iface_t));
+    return UCS_OK;
+
 }
 
 static UCS_F_ALWAYS_INLINE unsigned
@@ -150,11 +164,10 @@ uct_cuda_copy_queue_head_ready(ucs_queue_head_t *queue_head)
 
 static UCS_F_ALWAYS_INLINE unsigned
 uct_cuda_copy_progress_event_queue(uct_cuda_copy_iface_t *iface,
-                                   uct_cuda_copy_stream_t id,
+                                   ucs_queue_head_t *queue_head,
                                    unsigned max_events)
 {
-    ucs_queue_head_t *queue_head = &iface->outstanding_event_q[id];
-    unsigned count               = 0;
+    unsigned count = 0;
     uct_cuda_copy_event_desc_t *cuda_event;
 
     ucs_queue_for_each_extract(cuda_event, queue_head, queue,
@@ -171,6 +184,7 @@ uct_cuda_copy_progress_event_queue(uct_cuda_copy_iface_t *iface,
             break;
         }
     }
+
     return count;
 }
 
@@ -178,53 +192,64 @@ static unsigned uct_cuda_copy_iface_progress(uct_iface_h tl_iface)
 {
     uct_cuda_copy_iface_t *iface = ucs_derived_of(tl_iface, uct_cuda_copy_iface_t);
     unsigned max_events = iface->config.max_poll;
-    unsigned count;
+    unsigned count      = 0;
+    ucs_queue_head_t *event_q;
+    uct_cuda_copy_queue_desc_t *q_desc;
+    ucs_queue_iter_t iter;
+
+    ucs_queue_for_each_safe(q_desc, iter, &iface->active_queue, queue) {
+        event_q = &q_desc->event_queue;
+        count  += uct_cuda_copy_progress_event_queue(iface, event_q,
+                                                     max_events - count);
+        if (ucs_queue_is_empty(event_q)) {
+            ucs_queue_del_iter(&iface->active_queue, iter);
+        }
+    }
 
-    count = uct_cuda_copy_progress_event_queue(iface, UCT_CUDA_COPY_STREAM_D2H,
-                                               max_events);
-    count += uct_cuda_copy_progress_event_queue(iface, UCT_CUDA_COPY_STREAM_H2D,
-                                                (max_events - count));
     return count;
 }
 
-#if (__CUDACC_VER_MAJOR__ >= 100000)
-static void CUDA_CB myHostFn(void *cuda_copy_iface)
-#else
-static void CUDA_CB myHostCallback(CUstream hStream,  CUresult status,
-                                   void *cuda_copy_iface)
-#endif
-{
-    uct_cuda_copy_iface_t *iface = cuda_copy_iface;
-
-    ucs_assert(iface->async.event_cb != NULL);
-    /* notify user */
-    UCS_ASYNC_BLOCK(iface->super.worker->async);
-    iface->async.event_cb(iface->async.event_arg, 0);
-    UCS_ASYNC_UNBLOCK(iface->super.worker->async);
-}
-
 static ucs_status_t uct_cuda_copy_iface_event_fd_arm(uct_iface_h tl_iface,
                                                     unsigned events)
 {
     uct_cuda_copy_iface_t *iface = ucs_derived_of(tl_iface, uct_cuda_copy_iface_t);
-    int i;
     ucs_status_t status;
+    cudaStream_t *stream;
+    ucs_queue_head_t *event_q;
+    uct_cuda_copy_queue_desc_t *q_desc;
+    ucs_queue_iter_t iter;
 
-    for (i = 0; i < UCT_CUDA_COPY_STREAM_LAST; i++) {
-        if (uct_cuda_copy_queue_head_ready(&iface->outstanding_event_q[i])) {
+    ucs_queue_for_each_safe(q_desc, iter, &iface->active_queue, queue) {
+        event_q = &q_desc->event_queue;
+        if (uct_cuda_copy_queue_head_ready(event_q)) {
             return UCS_ERR_BUSY;
         }
     }
 
-    for (i = 0; i < UCT_CUDA_COPY_STREAM_LAST; i++) {
-        if (!ucs_queue_is_empty(&iface->outstanding_event_q[i])) {
+    status = ucs_async_eventfd_poll(iface->super.eventfd);
+    if (status == UCS_OK) {
+        return UCS_ERR_BUSY;
+    } else if (status == UCS_ERR_IO_ERROR) {
+        return status;
+    }
+
+    ucs_assertv(status == UCS_ERR_NO_PROGRESS, "%s", ucs_status_string(status));
+
+    ucs_queue_for_each_safe(q_desc, iter, &iface->active_queue, queue) {
+        event_q = &q_desc->event_queue;
+        stream  = &q_desc->stream;
+        if (!ucs_queue_is_empty(event_q)) {
             status =
 #if (__CUDACC_VER_MAJOR__ >= 100000)
-                UCT_CUDADRV_FUNC_LOG_ERR(cuLaunchHostFunc(iface->stream[i],
-                                         myHostFn, iface));
+                UCT_CUDADRV_FUNC_LOG_ERR(
+                        cuLaunchHostFunc(*stream,
+                                         uct_cuda_base_iface_stream_cb_fxn,
+                                         &iface->super));
 #else
-                UCT_CUDADRV_FUNC_LOG_ERR(cuStreamAddCallback(iface->stream[i],
-                                         myHostCallback, iface, 0));
+                UCT_CUDADRV_FUNC_LOG_ERR(
+                        cuStreamAddCallback(*stream,
+                                            uct_cuda_base_iface_stream_cb_fxn,
+                                            &iface->super, 0));
 #endif
             if (UCS_OK != status) {
                 return status;
@@ -251,7 +276,7 @@ static uct_iface_ops_t uct_cuda_copy_iface_ops = {
     .iface_progress_enable    = uct_base_iface_progress_enable,
     .iface_progress_disable   = uct_base_iface_progress_disable,
     .iface_progress           = uct_cuda_copy_iface_progress,
-    .iface_event_fd_get       = (uct_iface_event_fd_get_func_t)ucs_empty_function_return_success,
+    .iface_event_fd_get       = uct_cuda_base_iface_event_fd_get,
     .iface_event_arm          = uct_cuda_copy_iface_event_fd_arm,
     .iface_close              = UCS_CLASS_DELETE_FUNC_NAME(uct_cuda_copy_iface_t),
     .iface_query              = uct_cuda_copy_iface_query,
@@ -276,46 +301,79 @@ static void uct_cuda_copy_event_desc_init(ucs_mpool_t *mp, void *obj, void *chun
 static void uct_cuda_copy_event_desc_cleanup(ucs_mpool_t *mp, void *obj)
 {
     uct_cuda_copy_event_desc_t *base = (uct_cuda_copy_event_desc_t *) obj;
-    int active;
+    uct_cuda_copy_iface_t *iface     = ucs_container_of(mp,
+                                                        uct_cuda_copy_iface_t,
+                                                        cuda_event_desc);
+    CUcontext cuda_context;
 
-    UCT_CUDADRV_CTX_ACTIVE(active);
-
-    if (active) {
+    UCT_CUDADRV_FUNC_LOG_ERR(cuCtxGetCurrent(&cuda_context));
+    if (uct_cuda_base_context_match(cuda_context, iface->cuda_context)) {
         UCT_CUDA_FUNC_LOG_ERR(cudaEventDestroy(base->event));
     }
 }
 
 static ucs_status_t
-uct_cuda_copy_estimate_perf(uct_iface_h iface, uct_perf_attr_t *perf_attr)
+uct_cuda_copy_estimate_perf(uct_iface_h tl_iface, uct_perf_attr_t *perf_attr)
 {
+    uct_cuda_copy_iface_t *iface   = ucs_derived_of(tl_iface,
+                                                    uct_cuda_copy_iface_t);
+    uct_ep_operation_t op          = UCT_ATTR_VALUE(PERF, perf_attr, operation,
+                                                    OPERATION, UCT_EP_OP_LAST);
+    ucs_memory_type_t src_mem_type = UCT_ATTR_VALUE(PERF, perf_attr,
+                                                    local_memory_type,
+                                                    LOCAL_MEMORY_TYPE,
+                                                    UCS_MEMORY_TYPE_UNKNOWN);
+    ucs_memory_type_t dst_mem_type = UCT_ATTR_VALUE(PERF, perf_attr,
+                                                    remote_memory_type,
+                                                    REMOTE_MEMORY_TYPE,
+                                                    UCS_MEMORY_TYPE_UNKNOWN);
+    int zcopy                      = uct_ep_op_is_zcopy(op);
+    const double latency           = 1.8e-6;
+    const double overhead          = 4.0e-6;
+
     if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_BANDWIDTH) {
+        if (uct_ep_op_is_fetch(op)) {
+            ucs_swap(&src_mem_type, &dst_mem_type);
+        }
+
         perf_attr->bandwidth.dedicated = 0;
-        if (!(perf_attr->field_mask & UCT_PERF_ATTR_FIELD_OPERATION)) {
-            perf_attr->bandwidth.shared = UCT_CUDA_COPY_IFACE_DEFAULT_BANDWIDTH;
+        if ((src_mem_type == UCS_MEMORY_TYPE_HOST) &&
+            (dst_mem_type == UCS_MEMORY_TYPE_CUDA)) {
+            perf_attr->bandwidth.shared = (zcopy ? 8300.0 : 7900.0) * UCS_MBYTE;
+        } else if ((src_mem_type == UCS_MEMORY_TYPE_CUDA) &&
+                   (dst_mem_type == UCS_MEMORY_TYPE_HOST)) {
+            perf_attr->bandwidth.shared = (zcopy ? 11660.0 : 9320.0) *
+                                          UCS_MBYTE;
+        } else if ((src_mem_type == UCS_MEMORY_TYPE_CUDA) &&
+                   (dst_mem_type == UCS_MEMORY_TYPE_CUDA)) {
+            perf_attr->bandwidth.shared = 320.0 * UCS_GBYTE;
         } else {
-            switch (perf_attr->operation) {
-            case UCT_OP_GET_SHORT:
-                perf_attr->bandwidth.shared = 9320.0 * UCS_MBYTE;
-                break;
-            case UCT_OP_GET_ZCOPY:
-                perf_attr->bandwidth.shared = 11660.0 * UCS_MBYTE;
-                break;
-            case UCT_OP_PUT_SHORT:
-                perf_attr->bandwidth.shared = 8110.0 * UCS_MBYTE;
-                break;
-            case UCT_OP_PUT_ZCOPY:
-                perf_attr->bandwidth.shared = 9980.0 * UCS_MBYTE;
-                break;
-            default:
-                perf_attr->bandwidth.shared =
-                        UCT_CUDA_COPY_IFACE_DEFAULT_BANDWIDTH;
-                break;
-            }
+            perf_attr->bandwidth.shared = iface->config.bandwidth;
         }
     }
 
-    if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_OVERHEAD) {
-        perf_attr->overhead = UCT_CUDA_COPY_IFACE_OVERHEAD;
+    if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_SEND_PRE_OVERHEAD) {
+        perf_attr->send_pre_overhead = overhead;
+    }
+
+    if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_SEND_POST_OVERHEAD) {
+        /* In case of sync mem copy, the send operation CPU overhead includes
+           the latency of waiting for the copy to complete */
+        perf_attr->send_post_overhead = zcopy ? 0 : latency;
+    }
+
+    if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_RECV_OVERHEAD) {
+        perf_attr->recv_overhead = 0;
+    }
+
+    if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_LATENCY) {
+        /* In case of async mem copy, the latency is not part of the overhead
+           and it's a standalone property */
+        perf_attr->latency = ucs_linear_func_make(zcopy ? latency : 0.0, 0.0);
+    }
+
+    if (perf_attr->field_mask & UCT_PERF_ATTR_FIELD_MAX_INFLIGHT_EPS) {
+        perf_attr->max_inflight_eps = SIZE_MAX;
     }
 
     return UCS_OK;
@@ -326,11 +384,16 @@ static ucs_mpool_ops_t uct_cuda_copy_event_desc_mpool_ops = {
     .chunk_release = ucs_mpool_chunk_free,
     .obj_init      = uct_cuda_copy_event_desc_init,
     .obj_cleanup   = uct_cuda_copy_event_desc_cleanup,
+    .obj_str       = NULL
 };
 
 static uct_iface_internal_ops_t uct_cuda_copy_iface_internal_ops = {
-    .iface_estimate_perf = uct_cuda_copy_estimate_perf,
-    .iface_vfs_refresh   = (uct_iface_vfs_refresh_func_t)ucs_empty_function,
+    .iface_estimate_perf   = uct_cuda_copy_estimate_perf,
+    .iface_vfs_refresh     = (uct_iface_vfs_refresh_func_t)ucs_empty_function,
+    .ep_query              = (uct_ep_query_func_t)ucs_empty_function_return_unsupported,
+    .ep_invalidate         = (uct_ep_invalidate_func_t)ucs_empty_function_return_unsupported,
+    .ep_connect_to_ep_v2   = ucs_empty_function_return_unsupported,
+    .iface_is_reachable_v2 = uct_base_iface_is_reachable_v2
 };
 
 static UCS_CLASS_INIT_FUNC(uct_cuda_copy_iface_t, uct_md_h md, uct_worker_h worker,
@@ -339,14 +402,13 @@ static UCS_CLASS_INIT_FUNC(uct_cuda_copy_iface_t, uct_md_h md, uct_worker_h work
 {
     uct_cuda_copy_iface_config_t *config = ucs_derived_of(tl_config,
                                                           uct_cuda_copy_iface_config_t);
-    int i;
     ucs_status_t status;
+    ucs_memory_type_t src, dst;
+    ucs_mpool_params_t mp_params;
 
-    UCS_CLASS_CALL_SUPER_INIT(uct_base_iface_t, &uct_cuda_copy_iface_ops,
+    UCS_CLASS_CALL_SUPER_INIT(uct_cuda_iface_t, &uct_cuda_copy_iface_ops,
                               &uct_cuda_copy_iface_internal_ops, md, worker,
-                              params,
-                              tl_config UCS_STATS_ARG(params->stats_root)
-                              UCS_STATS_ARG("cuda_copy"));
+                              params, tl_config, "cuda_copy");
 
     if (strncmp(params->mode.device.dev_name,
                 UCT_CUDA_DEV_NAME, strlen(UCT_CUDA_DEV_NAME)) != 0) {
@@ -357,55 +419,74 @@ static UCS_CLASS_INIT_FUNC(uct_cuda_copy_iface_t, uct_md_h md, uct_worker_h work
     self->id                     = ucs_generate_uuid((uintptr_t)self);
     self->config.max_poll        = config->max_poll;
     self->config.max_cuda_events = config->max_cuda_events;
+    self->config.bandwidth       = config->bandwidth;
 
-    status = ucs_mpool_init(&self->cuda_event_desc,
-                            0,
-                            sizeof(uct_cuda_copy_event_desc_t),
-                            0,
-                            UCS_SYS_CACHE_LINE_SIZE,
-                            128,
-                            self->config.max_cuda_events,
-                            &uct_cuda_copy_event_desc_mpool_ops,
-                            "CUDA EVENT objects");
-
+    ucs_mpool_params_reset(&mp_params);
+    mp_params.elem_size       = sizeof(uct_cuda_copy_event_desc_t);
+    mp_params.elems_per_chunk = 128;
+    mp_params.max_elems       = self->config.max_cuda_events;
+    mp_params.ops             = &uct_cuda_copy_event_desc_mpool_ops;
+    mp_params.name            = "CUDA EVENT objects";
+    status = ucs_mpool_init(&mp_params, &self->cuda_event_desc);
     if (UCS_OK != status) {
         ucs_error("mpool creation failed");
         return UCS_ERR_IO_ERROR;
     }
 
-    uct_iface_set_async_event_params(params, &self->async.event_cb,
-                                     &self->async.event_arg);
+    ucs_queue_head_init(&self->active_queue);
 
-    for (i = 0; i < UCT_CUDA_COPY_STREAM_LAST; i++) {
-        self->stream[i] = 0;
-        ucs_queue_head_init(&self->outstanding_event_q[i]);
+    for (src = 0; src < UCS_MEMORY_TYPE_LAST; ++src) {
+        for (dst = 0; dst < UCS_MEMORY_TYPE_LAST; ++dst) {
+            self->queue_desc[src][dst].stream = 0;
+            ucs_queue_head_init(&self->queue_desc[src][dst].event_queue);
+        }
     }
 
+    self->short_stream = 0;
+    self->cuda_context = 0;
+
     return UCS_OK;
 }
 
 static UCS_CLASS_CLEANUP_FUNC(uct_cuda_copy_iface_t)
 {
-    int active;
-    int i;
+    cudaStream_t *stream;
+    CUcontext cuda_context;
+    ucs_queue_head_t *event_q;
+    ucs_memory_type_t src, dst;
 
-    UCT_CUDADRV_CTX_ACTIVE(active);
-
-    uct_base_iface_progress_disable(&self->super.super,
+    uct_base_iface_progress_disable(&self->super.super.super,
                                     UCT_PROGRESS_SEND | UCT_PROGRESS_RECV);
-    if (active) {
-        for (i = 0; i < UCT_CUDA_COPY_STREAM_LAST; i++) {
-            if (self->stream[i] != 0) {
-                ucs_assert(ucs_queue_is_empty(&self->outstanding_event_q[i]));
-                UCT_CUDA_FUNC_LOG_ERR(cudaStreamDestroy(self->stream[i]));
+
+    UCT_CUDADRV_FUNC_LOG_ERR(cuCtxGetCurrent(&cuda_context));
+    if (uct_cuda_base_context_match(cuda_context, self->cuda_context)) {
+
+        for (src = 0; src < UCS_MEMORY_TYPE_LAST; ++src) {
+            for (dst = 0; dst < UCS_MEMORY_TYPE_LAST; ++dst) {
+                stream  = &self->queue_desc[src][dst].stream;
+                event_q = &self->queue_desc[src][dst].event_queue;
+
+                if (!ucs_queue_is_empty(event_q)) {
+                    ucs_warn("stream destroyed but queue not empty");
+                }
+
+                if (*stream == 0) {
+                    continue;
+                }
+
+                UCT_CUDA_FUNC_LOG_ERR(cudaStreamDestroy(*stream));
             }
         }
+
+        if (self->short_stream) {
+            UCT_CUDA_FUNC_LOG_ERR(cudaStreamDestroy(self->short_stream));
+        }
     }
 
     ucs_mpool_cleanup(&self->cuda_event_desc, 1);
 }
 
-UCS_CLASS_DEFINE(uct_cuda_copy_iface_t, uct_base_iface_t);
+UCS_CLASS_DEFINE(uct_cuda_copy_iface_t, uct_cuda_iface_t);
 UCS_CLASS_DEFINE_NEW_FUNC(uct_cuda_copy_iface_t, uct_iface_t, uct_md_h, uct_worker_h,
                           const uct_iface_params_t*, const uct_iface_config_t*);
 static UCS_CLASS_DEFINE_DELETE_FUNC(uct_cuda_copy_iface_t, uct_iface_t);
