@@ -1,5 +1,5 @@
 /**
-* Copyright (C) Mellanox Technologies Ltd. 2001-2014.  ALL RIGHTS RESERVED.
+* Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2001-2014. ALL RIGHTS RESERVED.
 *
 * See file LICENSE for terms.
 */
@@ -87,9 +87,17 @@
     _desc->super.user_comp = _comp;
 
 
+#define UCT_RC_IFACE_CHECK_FLUSH_REMOTE(_condition, _ep, _iface, _name) \
+    if (ENABLE_PARAMS_CHECK && !(_condition)) { \
+        ucs_error("%s endpoint %p on %s: flush(remote) is not supported", \
+                  UCS_PP_QUOTE(_name), \
+                  _ep, \
+                  uct_ib_device_name(uct_ib_iface_device(&(_iface)->super))); \
+        return UCS_ERR_UNSUPPORTED; \
+    }
+
+
 enum {
-    UCT_RC_IFACE_STAT_RX_COMPLETION,
-    UCT_RC_IFACE_STAT_TX_COMPLETION,
     UCT_RC_IFACE_STAT_NO_CQE,
     UCT_RC_IFACE_STAT_NO_READS,
     UCT_RC_IFACE_STAT_LAST
@@ -147,9 +155,9 @@ typedef enum uct_rc_fence_mode {
 /* Common configuration used for rc verbs, rcx and dc transports */
 typedef struct uct_rc_iface_common_config {
     uct_ib_iface_config_t    super;
-    unsigned                 max_rd_atomic;
-    int                      ooo_rw; /* Enable out-of-order RDMA data placement */
+    unsigned long            max_rd_atomic;
     int                      fence_mode;
+    unsigned long            ece;
 
     struct {
         double               timeout;
@@ -176,8 +184,6 @@ struct uct_rc_iface_config {
     unsigned                       tx_cq_moderation; /* How many TX messages are
                                                         batched to one CQE */
     unsigned                       tx_cq_len;
-    unsigned                       log_ack_req_freq; /* Log of requests ack
-                                                        frequency on DevX */
 };
 
 
@@ -270,7 +276,7 @@ struct uct_rc_iface {
         unsigned             tx_qp_len;
         unsigned             tx_min_sge;
         unsigned             tx_min_inline;
-        unsigned             tx_ops_count;
+        unsigned             tx_cq_len;
         uint16_t             tx_moderation;
         uint8_t              tx_poll_always;
 
@@ -290,13 +296,10 @@ struct uct_rc_iface {
         uint8_t              rnr_retry;
         uint8_t              retry_cnt;
         uint8_t              max_rd_atomic;
-        /* Enable out-of-order RDMA data placement */
-        uint8_t              ooo_rw;
-#if UCS_ENABLE_ASSERT
-        int                  tx_cq_len;
-#endif
+        uint8_t              flush_remote;
         uct_rc_fence_mode_t  fence_mode;
         unsigned             exp_backoff;
+        uint32_t             ece;
         size_t               max_get_zcopy;
 
         /* Atomic callbacks */
@@ -376,8 +379,8 @@ ucs_status_t uct_rc_iface_query(uct_rc_iface_t *iface,
                                 size_t am_max_hdr, size_t am_max_iov,
                                 size_t am_min_hdr, size_t rma_max_iov);
 
-void uct_rc_iface_add_qp(uct_rc_iface_t *iface, uct_rc_ep_t *ep,
-                         unsigned qp_num);
+ucs_status_t
+uct_rc_iface_add_qp(uct_rc_iface_t *iface, uct_rc_ep_t *ep, unsigned qp_num);
 
 void uct_rc_iface_remove_qp(uct_rc_iface_t *iface, unsigned qp_num);
 
@@ -392,7 +395,6 @@ void uct_rc_iface_cleanup_qps(uct_rc_iface_t *iface);
 
 unsigned uct_rc_iface_qp_cleanup_progress(void *arg);
 
-
 /**
  * Creates an RC or DCI QP
  */
@@ -419,10 +421,10 @@ ucs_status_t uct_rc_iface_fc_handler(uct_rc_iface_t *iface, unsigned qp_num,
 ucs_status_t uct_rc_init_fc_thresh(uct_rc_iface_config_t *rc_cfg,
                                    uct_rc_iface_t *iface);
 
-ucs_status_t uct_rc_iface_event_arm(uct_iface_h tl_iface, unsigned events);
+uint64_t uct_rc_iface_arm_cq_check(uct_rc_iface_t *iface, unsigned events,
+                                   int *solicited_p);
 
-ucs_status_t uct_rc_iface_common_event_arm(uct_iface_h tl_iface,
-                                           unsigned events, int force_rx_all);
+ucs_status_t uct_rc_iface_event_arm(uct_iface_h tl_iface, unsigned events);
 
 ucs_status_t uct_rc_iface_init_rx(uct_rc_iface_t *iface,
                                   const uct_rc_iface_common_config_t *config,
@@ -430,6 +432,9 @@ ucs_status_t uct_rc_iface_init_rx(uct_rc_iface_t *iface,
 
 ucs_status_t uct_rc_iface_fence(uct_iface_h tl_iface, unsigned flags);
 
+ucs_status_t uct_rc_iface_estimate_perf(uct_iface_h tl_iface,
+                                        uct_perf_attr_t *perf_attr);
+
 void uct_rc_iface_vfs_populate(uct_rc_iface_t *iface);
 
 void uct_rc_iface_vfs_refresh(uct_iface_h iface);
@@ -438,6 +443,13 @@ ucs_arbiter_cb_result_t
 uct_rc_ep_process_pending(ucs_arbiter_t *arbiter, ucs_arbiter_group_t *group,
                           ucs_arbiter_elem_t *elem, void *arg);
 
+static UCS_F_ALWAYS_INLINE void
+uct_rc_iface_arbiter_dispatch(uct_rc_iface_t *iface)
+{
+    ucs_arbiter_dispatch(&iface->tx.arbiter, 1, uct_rc_ep_process_pending,
+                         NULL);
+}
+
 static UCS_F_ALWAYS_INLINE ucs_status_t
 uct_rc_fc_ctrl(uct_ep_t *ep, unsigned op, uct_rc_pending_req_t *req)
 {
@@ -479,21 +491,20 @@ uct_rc_iface_update_reads(uct_rc_iface_t *iface)
 }
 
 static UCS_F_ALWAYS_INLINE void
-uct_rc_iface_add_cq_credits_dispatch(uct_rc_iface_t *iface, uint16_t cq_credits)
+uct_rc_iface_add_cq_credits(uct_rc_iface_t *iface, uint16_t cq_credits)
 {
     iface->tx.cq_available += cq_credits;
-    ucs_assertv(iface->tx.cq_available <= iface->config.tx_cq_len,
-                "cq_available=%d tx_cq_len=%d cq_credits=%d",
+    ucs_assertv((ssize_t)iface->tx.cq_available <=
+                (ssize_t)iface->config.tx_cq_len,
+                "cq_available=%d tx_cq_len=%u cq_credits=%d",
                 iface->tx.cq_available, iface->config.tx_cq_len, cq_credits);
-    ucs_arbiter_dispatch(&iface->tx.arbiter, 1, uct_rc_ep_process_pending,
-                         NULL);
 }
 
 static UCS_F_ALWAYS_INLINE uct_rc_iface_send_op_t*
 uct_rc_iface_get_send_op(uct_rc_iface_t *iface)
 {
-    uct_rc_iface_send_op_t *op;
-    op = iface->tx.free_ops;
+    uct_rc_iface_send_op_t *op = iface->tx.free_ops;
+
     iface->tx.free_ops = op->next;
     return op;
 }
@@ -503,7 +514,8 @@ uct_rc_iface_put_send_op(uct_rc_iface_send_op_t *op)
 {
     uct_rc_iface_t *iface = op->iface;
 
-    ucs_assert(op->flags & UCT_RC_IFACE_SEND_OP_FLAG_IFACE);
+    ucs_assertv(op->flags == UCT_RC_IFACE_SEND_OP_FLAG_IFACE,
+                "op %p flags 0x%x", op, op->flags);
 
     op->next = iface->tx.free_ops;
     iface->tx.free_ops = op;
