@@ -1,5 +1,5 @@
 /**
-* Copyright (C) Mellanox Technologies Ltd. 2001-2016.  ALL RIGHTS RESERVED.
+* Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2001-2016. ALL RIGHTS RESERVED.
 * Copyright (C) Advanced Micro Devices, Inc. 2016 - 2017. ALL RIGHTS RESERVED.
 * See file LICENSE for terms.
 */
@@ -17,7 +17,11 @@ extern "C" {
 #include <ucs/arch/bitops.h>
 #include <ucs/arch/atomic.h>
 #include <ucs/sys/math.h>
+#if HAVE_IB
+#include <uct/ib/base/ib_md.h>
+#endif
 }
+#include <sys/resource.h>
 #include <net/if_arp.h>
 #include <ifaddrs.h>
 #include <netdb.h>
@@ -47,15 +51,100 @@ void* test_md::alloc_thread(void *arg)
 
     while (!*stop_flag) {
         int count = ucs::rand() % 100;
-        std::vector<void*> buffers;
+        ucs::ptr_vector<void> buffers;
         for (int i = 0; i < count; ++i) {
+            // allocate via malloc(), because ptr_vector<void>::release()
+            // method specialization uses free() to release a memory obtained
+            // for an element
             buffers.push_back(malloc(ucs::rand() % (256 * UCS_KBYTE)));
         }
-        std::for_each(buffers.begin(), buffers.end(), free);
     }
     return NULL;
 }
 
+ucs_status_t test_md::reg_mem(unsigned flags, void *address, size_t length,
+                              uct_mem_h *memh_p)
+{
+    uct_md_mem_reg_params_t reg_params;
+
+    reg_params.field_mask = UCT_MD_MEM_REG_FIELD_FLAGS;
+    reg_params.flags      = flags;
+
+    return uct_md_mem_reg_v2(md(), address, length, &reg_params, memh_p);
+}
+
+void test_md::test_reg_mem(unsigned access_mask,
+                           unsigned invalidate_flag)
+{
+    static const size_t size = 1 * UCS_MBYTE;
+
+    uct_mem_h memh;
+    void *ptr;
+    ucs_status_t status;
+    uct_md_mem_dereg_params_t params;
+
+    if ((access_mask & UCT_MD_MEM_ACCESS_REMOTE_ATOMIC) && is_bf_arm()) {
+        UCS_TEST_MESSAGE << "FIXME: AMO reg key bug on BF device, skipping";
+        return;
+    }
+
+    ptr    = malloc(size);
+    status = reg_mem(access_mask, ptr, size, &memh);
+    ASSERT_UCS_OK(status);
+
+    comp().comp.func   = dereg_cb;
+    comp().comp.count  = 1;
+    comp().comp.status = UCS_OK;
+    comp().self        = this;
+    params.memh        = memh;
+    params.flags       = UCT_MD_MEM_DEREG_FLAG_INVALIDATE;
+    params.comp        = &comp().comp;
+
+    if (!is_supported_reg_mem_flags(access_mask)) {
+        params.field_mask = UCT_MD_MEM_DEREG_FIELD_COMPLETION |
+                            UCT_MD_MEM_DEREG_FIELD_FLAGS |
+                            UCT_MD_MEM_DEREG_FIELD_MEMH;
+        status            = uct_md_mem_dereg_v2(md(), &params);
+        ASSERT_UCS_STATUS_EQ(UCS_ERR_UNSUPPORTED, status);
+
+        params.field_mask = UCT_MD_MEM_DEREG_FIELD_MEMH;
+        status            = uct_md_mem_dereg_v2(md(), &params);
+    } else {
+        params.field_mask = UCT_MD_MEM_DEREG_FIELD_COMPLETION;
+        status            = uct_md_mem_dereg_v2(md(), &params);
+        ASSERT_UCS_STATUS_EQ(UCS_ERR_INVALID_PARAM, status);
+
+        params.field_mask = UCT_MD_MEM_DEREG_FIELD_COMPLETION |
+                            UCT_MD_MEM_DEREG_FIELD_FLAGS;
+        status            = uct_md_mem_dereg_v2(md(), &params);
+        ASSERT_UCS_STATUS_EQ(UCS_ERR_INVALID_PARAM, status);
+
+        params.field_mask = UCT_MD_MEM_DEREG_FIELD_MEMH |
+                            UCT_MD_MEM_DEREG_FIELD_FLAGS;
+        status            = uct_md_mem_dereg_v2(md(), &params);
+        ASSERT_UCS_STATUS_EQ(UCS_ERR_INVALID_PARAM, status);
+
+        params.field_mask = UCT_MD_MEM_DEREG_FIELD_MEMH |
+                            UCT_MD_MEM_DEREG_FIELD_COMPLETION |
+                            UCT_MD_MEM_DEREG_FIELD_FLAGS;
+        status            = uct_md_mem_dereg_v2(md(), &params);
+        ASSERT_UCS_STATUS_EQ(UCS_ERR_INVALID_PARAM, status);
+
+        std::vector<uint8_t> rkey(md_attr().rkey_packed_size);
+        uct_md_mkey_pack_params_t pack_params;
+        pack_params.field_mask = UCT_MD_MKEY_PACK_FIELD_FLAGS;
+        pack_params.flags      = invalidate_flag;
+        status = uct_md_mkey_pack_v2(md(), memh, &pack_params, rkey.data());
+        EXPECT_UCS_OK(status);
+
+        status = uct_md_mem_dereg_v2(md(), &params);
+    }
+
+    EXPECT_UCS_OK(status);
+    free(ptr);
+}
+
+
 std::vector<test_md_param> test_md::enum_mds(const std::string& cmpt_name) {
 
     std::vector<md_resource> md_resources = enum_md_resources();
@@ -79,18 +168,42 @@ test_md::test_md()
                            (void (*)(uct_md_config_t*))uct_config_release,
                            uct_md_config_read, GetParam().component, NULL, NULL);
     memset(&m_md_attr, 0, sizeof(m_md_attr));
+    /* coverity[uninit_member] */
+}
+
+bool test_md::is_supported_reg_mem_flags(unsigned reg_flags) const
+{
+    return (reg_flags & md_flags_remote_rma) ?
+           check_caps(UCT_MD_FLAG_INVALIDATE_RMA) :
+           (reg_flags & UCT_MD_MEM_ACCESS_REMOTE_ATOMIC) ?
+           check_caps(UCT_MD_FLAG_INVALIDATE_AMO) : false;
+};
+
+bool test_md::is_bf_arm() const
+{
+    if ((ucs_arch_get_cpu_model() == UCS_CPU_MODEL_ARM_AARCH64) &&
+        (std::string("ib") == md_attr().component_name)) {
+#if HAVE_IB
+        uct_ib_md_t *ib_md = (uct_ib_md_t*)md();
+        if (ib_md->dev.pci_id.device == 0xa2d6) {
+            // BlueField 2
+            return true;
+        }
+#endif
+    }
+
+    return false;
 }
 
 void test_md::init()
 {
-    const std::vector<ucs_memory_type_t>
-        supported_mem_types = mem_buffer::supported_mem_types();
     ucs::test_base::init();
     UCS_TEST_CREATE_HANDLE(uct_md_h, m_md, uct_md_close, uct_md_open,
                            GetParam().component, GetParam().md_name.c_str(),
                            m_md_config);
 
-    ucs_status_t status = uct_md_query(m_md, &m_md_attr);
+    m_md_attr.field_mask = UINT64_MAX;
+    ucs_status_t status  = uct_md_query_v2(m_md, &m_md_attr);
     ASSERT_UCS_OK(status);
 
     check_skip_test();
@@ -113,15 +226,15 @@ void test_md::modify_config(const std::string& name, const std::string& value,
     }
 }
 
-bool test_md::check_caps(uint64_t flags)
+bool test_md::check_caps(uint64_t flags) const
 {
-    return ((md() == NULL) || ucs_test_all_flags(m_md_attr.cap.flags, flags));
+    return ((md() == NULL) || ucs_test_all_flags(m_md_attr.flags, flags));
 }
 
 bool test_md::check_reg_mem_type(ucs_memory_type_t mem_type)
 {
     return ((md() == NULL) || (check_caps(UCT_MD_FLAG_REG) &&
-                (m_md_attr.cap.reg_mem_types & UCS_BIT(mem_type))));
+                (m_md_attr.reg_mem_types & UCS_BIT(mem_type))));
 }
 
 void test_md::alloc_memory(void **address, size_t size, char *fill_buffer,
@@ -150,11 +263,11 @@ bool test_md::is_device_detected(ucs_memory_type_t mem_type)
            (mem_type != UCS_MEMORY_TYPE_ROCM_MANAGED);
 }
 
-void test_md::dereg_cb(void *arg)
+void test_md::dereg_cb(uct_completion_t *comp)
 {
-    test_md *self = (test_md*)arg;
+    test_md_comp_t *md_comp = ucs_container_of(comp, test_md_comp_t, comp);
 
-    self->m_comp_count++;
+    md_comp->self->m_comp_count++;
 }
 
 UCS_TEST_SKIP_COND_P(test_md, rkey_ptr,
@@ -261,7 +374,7 @@ UCS_TEST_SKIP_COND_P(test_md, alloc,
     params.mds.mds         = &md_ref;
     params.mds.count       = 1;
 
-    ucs_for_each_bit(mem_type, md_attr().cap.alloc_mem_types) {
+    ucs_for_each_bit(mem_type, md_attr().alloc_mem_types) {
         for (unsigned i = 0; i < 300; ++i) {
             size = orig_size = ucs::rand() % 65536;
             if (size == 0) {
@@ -299,11 +412,11 @@ UCS_TEST_P(test_md, mem_type_detect_mds) {
     int alloc_mem_type;
     void *address;
 
-    if (!md_attr().cap.detect_mem_types) {
+    if (!md_attr().detect_mem_types) {
         UCS_TEST_SKIP_R("MD can't detect any memory types");
     }
 
-    ucs_for_each_bit(alloc_mem_type, md_attr().cap.detect_mem_types) {
+    ucs_for_each_bit(alloc_mem_type, md_attr().detect_mem_types) {
         ucs_assert(alloc_mem_type < UCS_MEMORY_TYPE_LAST); /* for coverity */
 
         alloc_memory(&address, buffer_size, NULL,
@@ -337,8 +450,7 @@ UCS_TEST_P(test_md, mem_type_detect_mds) {
                                       slice_length, &mem_attr);
             ASSERT_UCS_OK(status);
             EXPECT_EQ(alloc_mem_type, mem_attr.mem_type);
-            if ((alloc_mem_type == UCS_MEMORY_TYPE_CUDA) ||
-                (alloc_mem_type == UCS_MEMORY_TYPE_CUDA_MANAGED)) {
+            if (alloc_mem_type == UCS_MEMORY_TYPE_CUDA) {
                 EXPECT_EQ(buffer_size, mem_attr.alloc_length);
                 EXPECT_EQ(address, mem_attr.base_address);
             } else {
@@ -349,24 +461,22 @@ UCS_TEST_P(test_md, mem_type_detect_mds) {
         }
 
         /* print memory type and dev name */
-        char sys_dev_name[128];
         mem_attr.field_mask = UCT_MD_MEM_ATTR_FIELD_SYS_DEV;
-
         status = uct_md_mem_query(md(), address, buffer_size, &mem_attr);
         ASSERT_UCS_OK(status);
 
-        ucs_topo_sys_device_bdf_name(mem_attr.sys_dev, sys_dev_name,
-                                     sizeof(sys_dev_name));
+        const char *dev_name = ucs_topo_sys_device_get_name(mem_attr.sys_dev);
         UCS_TEST_MESSAGE << ucs_memory_type_names[alloc_mem_type] << ": "
                          << "sys_dev[" << static_cast<int>(mem_attr.sys_dev)
-                         << "] (" << sys_dev_name << ")";
+                         << "] (" << dev_name << ")";
+
+        free_memory(address, static_cast<ucs_memory_type_t>(alloc_mem_type));
     }
 }
 
 UCS_TEST_P(test_md, mem_query) {
-    for (size_t i = 0; i < mem_buffer::supported_mem_types().size(); ++i) {
-        ucs_memory_type_t mem_type = mem_buffer::supported_mem_types()[i];
-        if (!(md_attr().cap.detect_mem_types & UCS_BIT(mem_type))) {
+    for (auto mem_type : mem_buffer::supported_mem_types()) {
+        if (!(md_attr().detect_mem_types & UCS_BIT(mem_type))) {
             continue;
         }
 
@@ -383,10 +493,8 @@ UCS_TEST_P(test_md, mem_query) {
             EXPECT_NE(UCS_SYS_DEVICE_ID_UNKNOWN, mem_attr.sys_dev);
         }
 
-        char bdf_buf[32];
         UCS_TEST_MESSAGE << ucs_memory_type_names[mem_type] << ": "
-                         << ucs_topo_sys_device_bdf_name(mem_attr.sys_dev, bdf_buf,
-                                                         sizeof(bdf_buf));
+                         << ucs_topo_sys_device_get_name(mem_attr.sys_dev);
     }
 }
 
@@ -399,12 +507,10 @@ UCS_TEST_P(test_md, sys_device) {
     ASSERT_UCS_OK(status);
 
     for (unsigned i = 0; i < num_tl_resources; ++i) {
-        char bdf_buf[32];
-        const char *bdf_name =
-                ucs_topo_sys_device_bdf_name(tl_resources[i].sys_device, bdf_buf,
-                                             sizeof(bdf_buf));
-        ASSERT_TRUE(bdf_name != NULL);
-        UCS_TEST_MESSAGE << tl_resources[i].dev_name << ": " << bdf_name;
+        const char *sysdev_name = ucs_topo_sys_device_get_name(
+                tl_resources[i].sys_device);
+        ASSERT_TRUE(sysdev_name != NULL);
+        UCS_TEST_MESSAGE << tl_resources[i].dev_name << ": " << sysdev_name;
 
         /* Expect 0 latency and infinite bandwidth within same device */
         ucs_sys_dev_distance_t distance;
@@ -432,7 +538,7 @@ UCS_TEST_SKIP_COND_P(test_md, reg,
 
     for (unsigned mem_type_id = 0; mem_type_id < UCS_MEMORY_TYPE_LAST; mem_type_id++) {
         ucs_memory_type_t mem_type = static_cast<ucs_memory_type_t>(mem_type_id);
-        if (!(md_attr().cap.reg_mem_types & UCS_BIT(mem_type_id))) {
+        if (!(md_attr().reg_mem_types & UCS_BIT(mem_type_id))) {
             UCS_TEST_MESSAGE << mem_buffer::mem_type_name(mem_type) << " memory "
                              << "registration is not supported by "
                              << GetParam().md_name;
@@ -474,7 +580,7 @@ UCS_TEST_SKIP_COND_P(test_md, reg_perf,
 
     for (unsigned mem_type_id = 0; mem_type_id < UCS_MEMORY_TYPE_LAST; mem_type_id++) {
         ucs_memory_type_t mem_type = static_cast<ucs_memory_type_t>(mem_type_id);
-        if (!(md_attr().cap.reg_mem_types & UCS_BIT(mem_type_id))) {
+        if (!(md_attr().reg_mem_types & UCS_BIT(mem_type_id))) {
             UCS_TEST_MESSAGE << mem_buffer::mem_type_name(mem_type) << " memory "
                              << " registration is not supported by "
                              << GetParam().md_name;
@@ -515,15 +621,23 @@ UCS_TEST_SKIP_COND_P(test_md, reg_perf,
     }
 }
 
-UCS_TEST_SKIP_COND_P(test_md, reg_advise,
-                     !check_caps(UCT_MD_FLAG_REG |
-                                 UCT_MD_FLAG_ADVISE)) {
-    size_t size;
+void test_md::test_reg_advise(size_t size, size_t advise_size,
+                              size_t advice_offset, bool check_non_blocking)
+{
+    ssize_t vmpin_before, vmpin_after;
     ucs_status_t status;
     void *address;
     uct_mem_h memh;
 
-    size = 128 * UCS_MBYTE;
+    if (check_non_blocking) {
+        if (!(md_attr().reg_nonblock_mem_types & UCS_BIT(UCS_MEMORY_TYPE_HOST))) {
+            UCS_TEST_SKIP_R("MD does not support non-blocking registration");
+        }
+
+        vmpin_before = ucs::get_proc_self_status_field("VmPin");
+        ASSERT_NE(vmpin_before, -1);
+    }
+
     address = malloc(size);
     ASSERT_TRUE(address != NULL);
 
@@ -532,16 +646,29 @@ UCS_TEST_SKIP_COND_P(test_md, reg_advise,
                             &memh);
     ASSERT_UCS_OK(status);
     ASSERT_TRUE(memh != UCT_MEM_HANDLE_NULL);
+    if (check_non_blocking) {
+        vmpin_after = ucs::get_proc_self_status_field("VmPin");
+        ASSERT_EQ(vmpin_before, vmpin_after);
+    }
 
-    status = uct_md_mem_advise(md(), memh, (char *)address + 7,
-                               32 * UCS_KBYTE, UCT_MADV_WILLNEED);
-    EXPECT_UCS_OK(status);
+    if (advise_size) {
+        status = uct_md_mem_advise(md(), memh,
+                                   UCS_PTR_BYTE_OFFSET(address, advice_offset),
+                                   advise_size, UCT_MADV_WILLNEED);
+        EXPECT_UCS_OK(status);
+    }
 
     status = uct_md_mem_dereg(md(), memh);
     EXPECT_UCS_OK(status);
     free(address);
 }
 
+UCS_TEST_SKIP_COND_P(test_md, reg_advise,
+                     !check_caps(UCT_MD_FLAG_REG | UCT_MD_FLAG_ADVISE))
+{
+    test_reg_advise(128 * UCS_MBYTE, 32 * UCS_KBYTE, 7);
+}
+
 UCS_TEST_SKIP_COND_P(test_md, alloc_advise,
                      !check_caps(UCT_MD_FLAG_ALLOC |
                                  UCT_MD_FLAG_ADVISE)) {
@@ -631,48 +758,73 @@ UCS_TEST_P(test_md, sockaddr_accessibility) {
     ASSERT_TRUE(getifaddrs(&ifaddr) != -1);
     /* go through a linked list of available interfaces */
     for (ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next) {
-        if (ucs::is_inet_addr(ifa->ifa_addr) &&
-            ucs_netif_flags_is_active(ifa->ifa_flags)) {
-            sock_addr.addr = ifa->ifa_addr;
-
-            UCS_TEST_MESSAGE << "Testing " << ifa->ifa_name << " with "
-                             << ucs::sockaddr_to_str(ifa->ifa_addr);
-            ASSERT_FALSE(uct_md_is_sockaddr_accessible(md(), &sock_addr,
-                                                       UCT_SOCKADDR_ACC_LOCAL));
-            ASSERT_FALSE(uct_md_is_sockaddr_accessible(md(), &sock_addr,
-                                                       UCT_SOCKADDR_ACC_REMOTE));
+        if (!ucs::is_interface_usable(ifa)) {
+            continue;
         }
+
+        sock_addr.addr = ifa->ifa_addr;
+        UCS_TEST_MESSAGE << "Testing " << ifa->ifa_name << " with "
+                         << ucs::sockaddr_to_str(ifa->ifa_addr);
+        ASSERT_FALSE(uct_md_is_sockaddr_accessible(md(), &sock_addr,
+                                                   UCT_SOCKADDR_ACC_LOCAL));
+        ASSERT_FALSE(uct_md_is_sockaddr_accessible(md(), &sock_addr,
+                                                   UCT_SOCKADDR_ACC_REMOTE));
     }
     freeifaddrs(ifaddr);
 }
 
 /* This test registers region N times and later deregs it N/2 times and
- * invalidates N/2 times - mix multiple dereg and invalidate calls */
+ * invalidates N/2 times - mix multiple dereg and invalidate calls.
+ * Guarantee that all packed keys are unique. */
 UCS_TEST_SKIP_COND_P(test_md, invalidate, !check_caps(UCT_MD_FLAG_INVALIDATE))
 {
-    static const size_t size = 1 * UCS_MBYTE;
+    static const size_t size       = 1 * UCS_MBYTE;
+    const int limit                = 2000 / ucs::test_time_multiplier();
+    static const unsigned md_flags = UCT_MD_MEM_ACCESS_REMOTE_PUT |
+                                     UCT_MD_MEM_ACCESS_REMOTE_GET;
     std::vector<uct_mem_h> memhs;
+    std::set<uint64_t> keys_set;
     uct_mem_h memh;
     void *ptr;
     size_t mem_reg_count; /* how many mem_reg operations to apply */
     size_t iter;
-    size_t comp_count;
     ucs_status_t status;
-    uct_md_mem_dereg_params_t params;
+    uct_md_mem_dereg_params_t dereg_params;
+    uct_md_mkey_pack_params_t pack_params;
+    uint64_t key;
 
-    params.field_mask = UCT_MD_MEM_DEREG_FIELD_FLAGS |
-                        UCT_MD_MEM_DEREG_FIELD_MEMH |
-                        UCT_MD_MEM_DEREG_FIELD_CALLBACK |
-                        UCT_MD_MEM_DEREG_FIELD_ARG;
-    params.cb         = dereg_cb;
-    params.arg        = this;
-    ptr               = malloc(size);
-    for (mem_reg_count = 1; mem_reg_count < 100; mem_reg_count++) {
-        comp_count   = (mem_reg_count + 1) / 2;
+    if (GetParam().md_name == "cuda_ipc") {
+        UCS_TEST_SKIP_R("test not needed with cuda-ipc");
+    }
+
+    comp().comp.func        = dereg_cb;
+    comp().comp.status      = UCS_OK;
+    comp().self             = this;
+    ptr                     = malloc(size);
+    dereg_params.field_mask = UCT_MD_MEM_DEREG_FIELD_FLAGS |
+                              UCT_MD_MEM_DEREG_FIELD_MEMH |
+                              UCT_MD_MEM_DEREG_FIELD_COMPLETION;
+    dereg_params.comp       = &comp().comp;
+
+    for (mem_reg_count = 1; mem_reg_count < limit; mem_reg_count++) {
+        comp().comp.count = (mem_reg_count + 1) / 2;
         m_comp_count = 0;
-        for (iter = 0; iter < mem_reg_count; iter++) {
-            status = uct_md_mem_reg(md(), ptr, size, UCT_MD_MEM_ACCESS_ALL,
-                                    &memh);
+
+        status = reg_mem(md_flags, ptr, size, &memh);
+        ASSERT_UCS_OK(status);
+        memhs.push_back(memh);
+
+        pack_params.field_mask = UCT_MD_MKEY_PACK_FIELD_FLAGS;
+        pack_params.flags      = UCT_MD_MKEY_PACK_FLAG_INVALIDATE_RMA;
+        status = uct_md_mkey_pack_v2(md(), memh, &pack_params, &key);
+        ASSERT_UCS_OK(status);
+
+        bool is_unique = keys_set.insert(key).second;
+        ASSERT_TRUE(is_unique) << keys_set.size()
+                               << "-th key is not unique";
+
+        for (iter = 1; iter < mem_reg_count; iter++) {
+            status = reg_mem(md_flags, ptr, size, &memh);
             ASSERT_UCS_OK(status);
             memhs.push_back(memh);
         }
@@ -682,79 +834,80 @@ UCS_TEST_SKIP_COND_P(test_md, invalidate, !check_caps(UCT_MD_FLAG_INVALIDATE))
             ASSERT_EQ(0, m_comp_count);
             memh = memhs.back();
             if ((iter & 1) == 0) { /* on even iteration invalidate handle */
-                params.flags = UCT_MD_MEM_DEREG_FLAG_INVALIDATE;
+                dereg_params.flags = UCT_MD_MEM_DEREG_FLAG_INVALIDATE;
             } else {
-                params.flags = 0;
+                dereg_params.flags = 0;
             }
 
-            params.memh = memh;
-            status      = uct_md_mem_dereg_v2(md(), &params);
+            dereg_params.memh = memh;
+            status            = uct_md_mem_dereg_v2(md(), &dereg_params);
             ASSERT_UCS_OK(status);
             memhs.pop_back();
         }
 
         ASSERT_TRUE(memhs.empty());
-        EXPECT_EQ(comp_count, m_comp_count);
+        EXPECT_EQ(1, m_comp_count);
     }
 
     free(ptr);
 }
 
-UCS_TEST_SKIP_COND_P(test_md, dereg_bad_arg,
-                     !check_reg_mem_type(UCS_MEMORY_TYPE_HOST))
+UCS_TEST_SKIP_COND_P(test_md, reg_bad_arg,
+                     !check_reg_mem_type(UCS_MEMORY_TYPE_HOST) ||
+                     !ENABLE_PARAMS_CHECK)
 {
-    static const size_t size = 1 * UCS_MBYTE;
     uct_mem_h memh;
-    void *ptr;
     ucs_status_t status;
-    uct_md_mem_dereg_params_t params;
 
-    ptr    = malloc(size);
-    status = uct_md_mem_reg(md(), ptr, size, UCT_MD_MEM_ACCESS_ALL, &memh);
+    status = reg_mem(UCT_MD_MEM_FLAG_HIDE_ERRORS, NULL, 0, &memh);
+    EXPECT_EQ(UCS_ERR_INVALID_PARAM, status);
+
+    status = reg_mem(UCT_MD_MEM_FLAG_HIDE_ERRORS | UCT_MD_MEM_FLAG_FIXED, NULL,
+                     0, &memh);
+    EXPECT_EQ(UCS_ERR_INVALID_PARAM, status);
+}
+
+UCS_TEST_SKIP_COND_P(test_md, dereg_bad_arg,
+                     !check_reg_mem_type(UCS_MEMORY_TYPE_HOST) ||
+                     !ENABLE_PARAMS_CHECK)
+{
+    test_reg_mem(md_flags_remote_rma, UCT_MD_MKEY_PACK_FLAG_INVALIDATE_RMA);
+    test_reg_mem(UCT_MD_MEM_ACCESS_REMOTE_ATOMIC,
+                 UCT_MD_MKEY_PACK_FLAG_INVALIDATE_RMA);
+}
+
+UCS_TEST_SKIP_COND_P(test_md, exported_mkey,
+                     !check_caps(UCT_MD_FLAG_EXPORTED_MKEY))
+{
+    size_t size   = ucs::rand() % UCS_MBYTE;
+    void *address = NULL;
+    uct_mem_h export_memh;
+    ucs_status_t status;
+
+    status = ucs_mmap_alloc(&size, &address, 0, "test_md_exp_mkey");
     ASSERT_UCS_OK(status);
 
-    params.cb    = dereg_cb;
-    params.arg   = this;
-    params.memh  = memh;
-    params.flags = UCT_MD_MEM_DEREG_FLAG_INVALIDATE;
+    UCS_TEST_MESSAGE << "allocated " << address << " of size " << size;
 
-    if (!check_caps(UCT_MD_FLAG_INVALIDATE)) {
-        params.field_mask = UCT_MD_MEM_DEREG_FIELD_CALLBACK |
-                            UCT_MD_MEM_DEREG_FIELD_ARG |
-                            UCT_MD_MEM_DEREG_FIELD_FLAGS |
-                            UCT_MD_MEM_DEREG_FIELD_MEMH;
-        status            = uct_md_mem_dereg_v2(md(), &params);
-        EXPECT_EQ(UCS_ERR_UNSUPPORTED, status);
+    status = reg_mem(UCT_MD_MEM_ACCESS_ALL, address, size, &export_memh);
+    ASSERT_UCS_OK(status);
 
-        params.field_mask = UCT_MD_MEM_DEREG_FIELD_MEMH;
-        status            = uct_md_mem_dereg_v2(md(), &params);
-    } else {
-        params.field_mask = UCT_MD_MEM_DEREG_FIELD_CALLBACK |
-                            UCT_MD_MEM_DEREG_FIELD_ARG;
-        status            = uct_md_mem_dereg_v2(md(), &params);
-        EXPECT_EQ(UCS_ERR_INVALID_PARAM, status);
+    std::vector<uint8_t> mkey_buffer(md_attr().exported_mkey_packed_size);
+    uct_md_mkey_pack_params_t pack_params;
+    pack_params.field_mask = UCT_MD_MKEY_PACK_FIELD_FLAGS;
+    pack_params.flags      = UCT_MD_MKEY_PACK_FLAG_EXPORT;
+    status = uct_md_mkey_pack_v2(md(), export_memh, &pack_params,
+                                 mkey_buffer.data());
+    ASSERT_UCS_OK(status);
 
-        params.field_mask = UCT_MD_MEM_DEREG_FIELD_CALLBACK |
-                            UCT_MD_MEM_DEREG_FIELD_ARG |
-                            UCT_MD_MEM_DEREG_FIELD_FLAGS;
-        status            = uct_md_mem_dereg_v2(md(), &params);
-        EXPECT_EQ(UCS_ERR_INVALID_PARAM, status);
+    uct_md_mem_dereg_params_t dereg_params;
+    dereg_params.field_mask = UCT_MD_MEM_DEREG_FIELD_MEMH;
+    dereg_params.memh       = export_memh;
+    status                  = uct_md_mem_dereg_v2(md(), &dereg_params);
+    ASSERT_UCS_OK(status);
 
-        params.field_mask = UCT_MD_MEM_DEREG_FIELD_MEMH |
-                            UCT_MD_MEM_DEREG_FIELD_ARG |
-                            UCT_MD_MEM_DEREG_FIELD_FLAGS;
-        status            = uct_md_mem_dereg_v2(md(), &params);
-        EXPECT_EQ(UCS_ERR_INVALID_PARAM, status);
-
-        params.field_mask = UCT_MD_MEM_DEREG_FIELD_MEMH |
-                            UCT_MD_MEM_DEREG_FIELD_CALLBACK |
-                            UCT_MD_MEM_DEREG_FIELD_ARG |
-                            UCT_MD_MEM_DEREG_FIELD_FLAGS;
-        status            = uct_md_mem_dereg_v2(md(), &params);
-    }
-
-    EXPECT_UCS_OK(status);
-    free(ptr);
+    status = ucs_mmap_free(address, size);
+    ASSERT_UCS_OK(status);
 }
 
 UCT_MD_INSTANTIATE_TEST_CASE(test_md)
@@ -821,3 +974,81 @@ UCS_TEST_SKIP_COND_P(test_md_fork, fork,
 
 UCT_MD_INSTANTIATE_TEST_CASE(test_md_fork)
 
+class test_md_memlock_limit : public test_md {
+protected:
+    void init() override
+    {
+        ucs::test_base::init();
+        check_skip_test();
+
+        if (getrlimit(RLIMIT_MEMLOCK, &m_previous_limit) != 0) {
+            UCS_TEST_SKIP_R("Cannot get the previous memlock limit");
+        }
+        const struct rlimit new_limit = {1, m_previous_limit.rlim_max};
+        if (setrlimit(RLIMIT_MEMLOCK, &new_limit) != 0) {
+            UCS_TEST_SKIP_R("Cannot set the new memlock limit");
+        }
+    }
+
+    void cleanup() override
+    {
+        if (setrlimit(RLIMIT_MEMLOCK, &m_previous_limit) != 0) {
+            UCS_TEST_ABORT("Failed to restore memlock limit. "
+                           "Can cause other tests failures!");
+        }
+        test_md::cleanup();
+    }
+
+    struct rlimit m_previous_limit;
+};
+
+UCS_TEST_P(test_md_memlock_limit, md_open)
+{
+    UCS_TEST_CREATE_HANDLE(uct_md_h, m_md, uct_md_close, uct_md_open,
+                           GetParam().component, GetParam().md_name.c_str(),
+                           m_md_config);
+}
+
+UCT_MD_INSTANTIATE_TEST_CASE(test_md_memlock_limit)
+
+class test_md_non_blocking : public test_md
+{
+protected:
+    void init() override {
+        /* ODPv1 IB feature can work only for certain DEVX configuration */
+        modify_config("MLX5_DEVX_OBJECTS", "dct,dcsrq", IGNORE_IF_NOT_EXIST);
+        test_md::init();
+    }
+};
+
+UCS_TEST_SKIP_COND_P(test_md_non_blocking, reg_advise,
+                     !check_caps(UCT_MD_FLAG_REG | UCT_MD_FLAG_ADVISE))
+{
+    test_reg_advise(UCS_KBYTE, UCS_KBYTE, 0, true);
+    test_reg_advise(UCS_KBYTE, UCS_KBYTE / 2, 0, true);
+    test_reg_advise(UCS_KBYTE, UCS_KBYTE / 2, UCS_KBYTE / 4, true);
+
+    /*
+     * TODO: These tests should be enabled
+     * when https://redmine.mellanox.com/issues/336376 fixed
+
+     * test_reg_advise(UCS_MBYTE, UCS_MBYTE, 0, true);
+     * test_reg_advise(UCS_MBYTE, UCS_MBYTE / 2, 0, true);
+     * test_reg_advise(UCS_MBYTE, UCS_MBYTE / 2, UCS_MBYTE / 4, true);
+     */
+}
+
+UCS_TEST_SKIP_COND_P(test_md_non_blocking, reg,
+                     !check_caps(UCT_MD_FLAG_REG))
+{
+    test_reg_advise(UCS_KBYTE, 0, 0, true);
+
+    /*
+     * TODO: This test should be enabled
+     * when https://redmine.mellanox.com/issues/336376 fixed
+
+     * test_reg_advise(UCS_MBYTE, 0, 0, true);
+     */
+}
+
+UCT_MD_INSTANTIATE_TEST_CASE(test_md_non_blocking)
