@@ -18,10 +18,15 @@
         }                                                               \
     } while (0)
 
+typedef struct cudai_stream_s {
+    bool created;
+    cudaStream_t stream;
+} cudai_stream;
+
 typedef struct {
     int ndevices;
-    cudaStream_t *stream;
-    bool **p2p;
+    cudai_stream *streams;      /* array of lazily created streams, one for each device */
+    int **p2p;                  /* p2p[sdev][ddev] */
 } yaksuri_cudai_global_s;
 extern yaksuri_cudai_global_s yaksuri_cudai_global;
 
