@@ -161,23 +161,25 @@ def populate_pupfns(pup_max_nesting, backend, blklens, builtin_types, builtin_ma
         yutils.display(OUTFILE, "\n")
 
     yutils.display(OUTFILE, "switch (type->kind) {\n")
-    for dtype1 in derived_types:
-        yutils.display(OUTFILE, "case YAKSI_TYPE_KIND__%s:\n" % dtype1.upper())
-        yutils.display(OUTFILE, "switch (type->u.%s.child->kind) {\n" % dtype1)
-        for dtype2 in derived_types:
-            yutils.display(OUTFILE, "case YAKSI_TYPE_KIND__%s:\n" % dtype2.upper())
-            yutils.display(OUTFILE, "rc = yaksuri_%si_populate_pupfns_%s_%s(type);\n" % (backend, dtype1, dtype2))
+    if pup_max_nesting > 0:
+        for dtype1 in derived_types:
+            yutils.display(OUTFILE, "case YAKSI_TYPE_KIND__%s:\n" % dtype1.upper())
+            yutils.display(OUTFILE, "switch (type->u.%s.child->kind) {\n" % dtype1)
+            if pup_max_nesting > 1:
+                for dtype2 in derived_types:
+                    yutils.display(OUTFILE, "case YAKSI_TYPE_KIND__%s:\n" % dtype2.upper())
+                    yutils.display(OUTFILE, "rc = yaksuri_%si_populate_pupfns_%s_%s(type);\n" % (backend, dtype1, dtype2))
+                    yutils.display(OUTFILE, "break;\n")
+                    yutils.display(OUTFILE, "\n")
+            yutils.display(OUTFILE, "case YAKSI_TYPE_KIND__BUILTIN:\n")
+            yutils.display(OUTFILE, "rc = yaksuri_%si_populate_pupfns_%s_builtin(type);\n" % (backend, dtype1))
+            yutils.display(OUTFILE, "break;\n")
+            yutils.display(OUTFILE, "\n")
+            yutils.display(OUTFILE, "default:\n")
+            yutils.display(OUTFILE, "    break;\n")
+            yutils.display(OUTFILE, "}\n")
             yutils.display(OUTFILE, "break;\n")
             yutils.display(OUTFILE, "\n")
-        yutils.display(OUTFILE, "case YAKSI_TYPE_KIND__BUILTIN:\n")
-        yutils.display(OUTFILE, "rc = yaksuri_%si_populate_pupfns_%s_builtin(type);\n" % (backend, dtype1))
-        yutils.display(OUTFILE, "break;\n")
-        yutils.display(OUTFILE, "\n")
-        yutils.display(OUTFILE, "default:\n")
-        yutils.display(OUTFILE, "    break;\n")
-        yutils.display(OUTFILE, "}\n")
-        yutils.display(OUTFILE, "break;\n")
-        yutils.display(OUTFILE, "\n")
     yutils.display(OUTFILE, "case YAKSI_TYPE_KIND__BUILTIN:\n")
     yutils.display(OUTFILE, "rc = yaksuri_%si_populate_pupfns_builtin(type);\n" % backend)
     yutils.display(OUTFILE, "break;\n")
@@ -190,9 +192,46 @@ def populate_pupfns(pup_max_nesting, backend, blklens, builtin_types, builtin_ma
     yutils.display(OUTFILE, "}\n");
     OUTFILE.close()
 
-    for dtype1 in derived_types:
-        for dtype2 in derived_types:
-            filename = "src/backend/%s/pup/yaksuri_%si_populate_pupfns_%s_%s.c" % (backend, backend, dtype1, dtype2)
+    if pup_max_nesting > 0:
+        for dtype1 in derived_types:
+            if pup_max_nesting > 1:
+                for dtype2 in derived_types:
+                    filename = "src/backend/%s/pup/yaksuri_%si_populate_pupfns_%s_%s.c" % (backend, backend, dtype1, dtype2)
+                    yutils.copyright_c(filename)
+                    OUTFILE = open(filename, "a")
+                    yutils.display(OUTFILE, "#include <stdio.h>\n")
+                    yutils.display(OUTFILE, "#include <stdlib.h>\n")
+                    yutils.display(OUTFILE, "#include <wchar.h>\n")
+                    yutils.display(OUTFILE, "#include \"yaksi.h\"\n")
+                    yutils.display(OUTFILE, "#include \"yaksu.h\"\n")
+                    yutils.display(OUTFILE, "#include \"yaksuri_%si.h\"\n" % backend)
+                    yutils.display(OUTFILE, "#include \"yaksuri_%si_populate_pupfns.h\"\n" % backend)
+                    yutils.display(OUTFILE, "#include \"yaksuri_%si_pup.h\"\n" % backend)
+                    yutils.display(OUTFILE, "\n")
+                    yutils.display(OUTFILE, "int yaksuri_%si_populate_pupfns_%s_%s(yaksi_type_s * type)\n" % (backend, dtype1, dtype2))
+                    yutils.display(OUTFILE, "{\n")
+                    yutils.display(OUTFILE, "int rc = YAKSA_SUCCESS;\n")
+                    yutils.display(OUTFILE, "yaksuri_%si_type_s *%s = (yaksuri_%si_type_s *) type->backend.%s.priv;\n" \
+                                   % (backend, backend, backend, backend))
+                    yutils.display(OUTFILE, "\n")
+                    yutils.display(OUTFILE, "char *str = getenv(\"YAKSA_ENV_MAX_NESTING_LEVEL\");\n")
+                    yutils.display(OUTFILE, "int max_nesting_level;\n")
+                    yutils.display(OUTFILE, "if (str) {\n")
+                    yutils.display(OUTFILE, "max_nesting_level = atoi(str);\n")
+                    yutils.display(OUTFILE, "} else {\n")
+                    yutils.display(OUTFILE, "max_nesting_level = YAKSI_ENV_DEFAULT_NESTING_LEVEL;\n")
+                    yutils.display(OUTFILE, "}\n")
+                    yutils.display(OUTFILE, "\n")
+
+                    pupstr = "%s_%s" % (dtype1, dtype2)
+                    typelist = [ dtype1, dtype2 ]
+                    switcher(backend, OUTFILE, blklens, builtin_types, builtin_maps, typelist, pupstr, pup_max_nesting - 1)
+                    yutils.display(OUTFILE, "\n")
+                    yutils.display(OUTFILE, "return rc;\n")
+                    yutils.display(OUTFILE, "}\n")
+                    OUTFILE.close()
+
+            filename = "src/backend/%s/pup/yaksuri_%si_populate_pupfns_%s_builtin.c" % (backend, backend, dtype1)
             yutils.copyright_c(filename)
             OUTFILE = open(filename, "a")
             yutils.display(OUTFILE, "#include <stdio.h>\n")
@@ -204,7 +243,7 @@ def populate_pupfns(pup_max_nesting, backend, blklens, builtin_types, builtin_ma
             yutils.display(OUTFILE, "#include \"yaksuri_%si_populate_pupfns.h\"\n" % backend)
             yutils.display(OUTFILE, "#include \"yaksuri_%si_pup.h\"\n" % backend)
             yutils.display(OUTFILE, "\n")
-            yutils.display(OUTFILE, "int yaksuri_%si_populate_pupfns_%s_%s(yaksi_type_s * type)\n" % (backend, dtype1, dtype2))
+            yutils.display(OUTFILE, "int yaksuri_%si_populate_pupfns_%s_builtin(yaksi_type_s * type)\n" % (backend, dtype1))
             yutils.display(OUTFILE, "{\n")
             yutils.display(OUTFILE, "int rc = YAKSA_SUCCESS;\n")
             yutils.display(OUTFILE, "yaksuri_%si_type_s *%s = (yaksuri_%si_type_s *) type->backend.%s.priv;\n" \
@@ -219,49 +258,14 @@ def populate_pupfns(pup_max_nesting, backend, blklens, builtin_types, builtin_ma
             yutils.display(OUTFILE, "}\n")
             yutils.display(OUTFILE, "\n")
 
-            pupstr = "%s_%s" % (dtype1, dtype2)
-            typelist = [ dtype1, dtype2 ]
-            switcher(backend, OUTFILE, blklens, builtin_types, builtin_maps, typelist, pupstr, pup_max_nesting - 1)
+            pupstr = "%s" % dtype1
+            typelist = [ dtype1 ]
+            switcher_builtin(backend, OUTFILE, blklens, builtin_types, builtin_maps, typelist, pupstr)
             yutils.display(OUTFILE, "\n")
             yutils.display(OUTFILE, "return rc;\n")
             yutils.display(OUTFILE, "}\n")
             OUTFILE.close()
 
-        filename = "src/backend/%s/pup/yaksuri_%si_populate_pupfns_%s_builtin.c" % (backend, backend, dtype1)
-        yutils.copyright_c(filename)
-        OUTFILE = open(filename, "a")
-        yutils.display(OUTFILE, "#include <stdio.h>\n")
-        yutils.display(OUTFILE, "#include <stdlib.h>\n")
-        yutils.display(OUTFILE, "#include <wchar.h>\n")
-        yutils.display(OUTFILE, "#include \"yaksi.h\"\n")
-        yutils.display(OUTFILE, "#include \"yaksu.h\"\n")
-        yutils.display(OUTFILE, "#include \"yaksuri_%si.h\"\n" % backend)
-        yutils.display(OUTFILE, "#include \"yaksuri_%si_populate_pupfns.h\"\n" % backend)
-        yutils.display(OUTFILE, "#include \"yaksuri_%si_pup.h\"\n" % backend)
-        yutils.display(OUTFILE, "\n")
-        yutils.display(OUTFILE, "int yaksuri_%si_populate_pupfns_%s_builtin(yaksi_type_s * type)\n" % (backend, dtype1))
-        yutils.display(OUTFILE, "{\n")
-        yutils.display(OUTFILE, "int rc = YAKSA_SUCCESS;\n")
-        yutils.display(OUTFILE, "yaksuri_%si_type_s *%s = (yaksuri_%si_type_s *) type->backend.%s.priv;\n" \
-                       % (backend, backend, backend, backend))
-        yutils.display(OUTFILE, "\n")
-        yutils.display(OUTFILE, "char *str = getenv(\"YAKSA_ENV_MAX_NESTING_LEVEL\");\n")
-        yutils.display(OUTFILE, "int max_nesting_level;\n")
-        yutils.display(OUTFILE, "if (str) {\n")
-        yutils.display(OUTFILE, "max_nesting_level = atoi(str);\n")
-        yutils.display(OUTFILE, "} else {\n")
-        yutils.display(OUTFILE, "max_nesting_level = YAKSI_ENV_DEFAULT_NESTING_LEVEL;\n")
-        yutils.display(OUTFILE, "}\n")
-        yutils.display(OUTFILE, "\n")
-
-        pupstr = "%s" % dtype1
-        typelist = [ dtype1 ]
-        switcher_builtin(backend, OUTFILE, blklens, builtin_types, builtin_maps, typelist, pupstr)
-        yutils.display(OUTFILE, "\n")
-        yutils.display(OUTFILE, "return rc;\n")
-        yutils.display(OUTFILE, "}\n")
-        OUTFILE.close()
-
     ### Generate mapping for reduction kernels of contiguous datatypes
     filename = "src/backend/%s/pup/yaksuri_%si_populate_pupfns_builtin.c" % (backend, backend)
     yutils.copyright_c(filename)
@@ -295,10 +299,12 @@ def populate_pupfns(pup_max_nesting, backend, blklens, builtin_types, builtin_ma
     yutils.copyright_makefile(filename)
     OUTFILE = open(filename, "a")
     yutils.display(OUTFILE, "libyaksa_la_SOURCES += \\\n")
-    for dtype1 in derived_types:
-        for dtype2 in derived_types:
-            yutils.display(OUTFILE, "\tsrc/backend/%s/pup/yaksuri_%si_populate_pupfns_%s_%s.c \\\n" % (backend, backend, dtype1, dtype2))
-        yutils.display(OUTFILE, "\tsrc/backend/%s/pup/yaksuri_%si_populate_pupfns_%s_builtin.c \\\n" % (backend, backend, dtype1))
+    if pup_max_nesting > 0:
+        for dtype1 in derived_types:
+            if pup_max_nesting > 1:
+                for dtype2 in derived_types:
+                    yutils.display(OUTFILE, "\tsrc/backend/%s/pup/yaksuri_%si_populate_pupfns_%s_%s.c \\\n" % (backend, backend, dtype1, dtype2))
+            yutils.display(OUTFILE, "\tsrc/backend/%s/pup/yaksuri_%si_populate_pupfns_%s_builtin.c \\\n" % (backend, backend, dtype1))
     yutils.display(OUTFILE, "\tsrc/backend/%s/pup/yaksuri_%si_populate_pupfns_builtin.c \\\n" % (backend, backend))
     yutils.display(OUTFILE, "\tsrc/backend/%s/pup/yaksuri_%si_populate_pupfns.c\n" % (backend, backend))
     yutils.display(OUTFILE, "\n")
@@ -313,10 +319,12 @@ def populate_pupfns(pup_max_nesting, backend, blklens, builtin_types, builtin_ma
     yutils.display(OUTFILE, "#ifndef YAKSURI_%sI_POPULATE_PUPFNS_H_INCLUDED\n" % backend.upper())
     yutils.display(OUTFILE, "#define YAKSURI_%sI_POPULATE_PUPFNS_H_INCLUDED\n" % backend.upper())
     yutils.display(OUTFILE, "\n")
-    for dtype1 in derived_types:
-        for dtype2 in derived_types:
-            yutils.display(OUTFILE, "int yaksuri_%si_populate_pupfns_%s_%s(yaksi_type_s * type);\n" % (backend, dtype1, dtype2))
-        yutils.display(OUTFILE, "int yaksuri_%si_populate_pupfns_%s_builtin(yaksi_type_s * type);\n" % (backend, dtype1))
+    if pup_max_nesting > 0:
+        for dtype1 in derived_types:
+            if pup_max_nesting > 1:
+                for dtype2 in derived_types:
+                    yutils.display(OUTFILE, "int yaksuri_%si_populate_pupfns_%s_%s(yaksi_type_s * type);\n" % (backend, dtype1, dtype2))
+            yutils.display(OUTFILE, "int yaksuri_%si_populate_pupfns_%s_builtin(yaksi_type_s * type);\n" % (backend, dtype1))
     yutils.display(OUTFILE, "int yaksuri_%si_populate_pupfns_builtin(yaksi_type_s * type);\n" % backend)
     yutils.display(OUTFILE, "\n")
     yutils.display(OUTFILE, "#endif  /* YAKSURI_%sI_POPULATE_PUPFNS_H_INCLUDED */\n" % backend.upper())
