@@ -5,12 +5,7 @@
 
 #include "cdesc.h"
 #include <limits.h>
-
-MPI_Status *MPIR_C_MPI_STATUS_IGNORE = MPI_STATUS_IGNORE;
-MPI_Status *MPIR_C_MPI_STATUSES_IGNORE = MPI_STATUSES_IGNORE;
-char **MPIR_C_MPI_ARGV_NULL = MPI_ARGV_NULL;
-char ***MPIR_C_MPI_ARGVS_NULL = MPI_ARGVS_NULL;
-int *MPIR_C_MPI_ERRCODES_IGNORE = MPI_ERRCODES_IGNORE;
+#include <assert.h>
 
 /* Fortran 2008 specifies a maximum rank of 15 */
 #define MAX_RANK  (15)
@@ -30,12 +25,12 @@ int cdesc_create_datatype(CFI_cdesc_t * cdesc, MPI_Aint oldcount, MPI_Datatype o
 #ifdef HAVE_ERROR_CHECKING
     {
         int size;
-        MPIR_Assert(cdesc->rank <= MAX_RANK);
+        assert(cdesc->rank <= MAX_RANK);
         PMPI_Type_size(oldtype, &size);
         /* When cdesc->elem_len != size, things suddenly become complicated. Generally, it is hard to create
          * a composite datatype based on two datatypes. Currently we don't support it and doubt it is useful.
          */
-        MPIR_Assert(cdesc->elem_len == size);
+        assert(cdesc->elem_len == size);
     }
 #endif
 
