@@ -17,11 +17,8 @@ int MPIR_Ireduce_scatter_block_intra_sched_noncommutative(const void *sendbuf, v
     int mpi_errno = MPI_SUCCESS;
     int comm_size = comm_ptr->local_size;
     int rank = comm_ptr->rank;
-    int pof2;
     int log2_comm_size;
     int i, k;
-    int recv_offset, send_offset;
-    int block_size, total_count, size;
     MPI_Aint true_extent, true_lb;
     int buf0_was_inout;
     void *tmp_buf0;
@@ -30,20 +27,16 @@ int MPIR_Ireduce_scatter_block_intra_sched_noncommutative(const void *sendbuf, v
 
     MPIR_Type_get_true_extent_impl(datatype, &true_lb, &true_extent);
 
-    pof2 = 1;
-    log2_comm_size = 0;
-    while (pof2 < comm_size) {
-        pof2 <<= 1;
-        ++log2_comm_size;
-    }
-
 #ifdef HAVE_ERROR_CHECKING
     /* begin error checking */
-    MPIR_Assert(pof2 == comm_size);     /* FIXME this version only works for power of 2 procs */
+    MPIR_Assert(MPL_pof2(comm_size));   /* FIXME this version only works for power of 2 procs */
     /* end error checking */
 #endif
 
+    log2_comm_size = MPL_log2(comm_size);
+
     /* size of a block (count of datatype per block, NOT bytes per block) */
+    MPI_Aint block_size, total_count;
     block_size = recvcount;
     total_count = block_size * comm_size;
 
@@ -71,6 +64,7 @@ int MPIR_Ireduce_scatter_block_intra_sched_noncommutative(const void *sendbuf, v
     MPIR_SCHED_BARRIER(s);
     buf0_was_inout = 1;
 
+    MPI_Aint send_offset, recv_offset, size;
     send_offset = 0;
     recv_offset = 0;
     size = total_count;
