@@ -9,9 +9,6 @@
 
 /* This is the utility file for info that contains the basic info items
    and storage management */
-#ifndef MPIR_INFO_PREALLOC
-#define MPIR_INFO_PREALLOC 8
-#endif
 
 /* Preallocated info objects */
 MPIR_Info MPIR_Info_builtin[MPIR_INFO_N_BUILTIN];
@@ -27,30 +24,29 @@ MPIR_Object_alloc_t MPIR_Info_mem = { 0, 0, 0, 0, 0, 0, MPIR_INFO,
    relies on the SINGLE_CS in the info routines (particularly MPI_Info_free) */
 int MPIR_Info_free_impl(MPIR_Info * info_ptr)
 {
-    MPIR_Info *curr_ptr, *last_ptr;
-
-    curr_ptr = info_ptr->next;
-    last_ptr = NULL;
-
-    MPIR_Info_handle_obj_free(&MPIR_Info_mem, info_ptr);
-
-    /* printf("Returning info %x\n", info_ptr->id); */
-    /* First, free the string storage */
-    while (curr_ptr) {
+    for (int i = 0; i < info_ptr->size; i++) {
         /* MPI_Info objects are allocated by normal MPL_direct_xxx() functions, so
          * they need to be freed by MPL_direct_free(), not MPL_free(). */
-        MPL_direct_free(curr_ptr->key);
-        MPL_direct_free(curr_ptr->value);
-        last_ptr = curr_ptr;
-        curr_ptr = curr_ptr->next;
-        MPIR_Info_handle_obj_free(&MPIR_Info_mem, last_ptr);
+        MPL_direct_free(info_ptr->entries[i].key);
+        MPL_direct_free(info_ptr->entries[i].value);
     }
+    MPL_direct_free(info_ptr->entries);
+    if (!HANDLE_IS_BUILTIN(info_ptr->handle)) {
+        MPIR_Info_handle_obj_free(&MPIR_Info_mem, info_ptr);
+    }
+
     return MPI_SUCCESS;
 }
 
-/* Allocate and initialize an MPIR_Info object.
- *
- * Returns MPICH error codes */
+/* Allocate and initialize an MPIR_Info object. */
+
+static void info_init(MPIR_Info * info_ptr)
+{
+    info_ptr->capacity = 0;
+    info_ptr->size = 0;
+    info_ptr->entries = NULL;
+}
+
 int MPIR_Info_alloc(MPIR_Info ** info_p_p)
 {
     int mpi_errno = MPI_SUCCESS;
@@ -58,9 +54,7 @@ int MPIR_Info_alloc(MPIR_Info ** info_p_p)
     MPIR_ERR_CHKANDJUMP1(!*info_p_p, mpi_errno, MPI_ERR_OTHER, "**nomem", "**nomem %s", "MPI_Info");
 
     MPIR_Object_set_ref(*info_p_p, 0);
-    (*info_p_p)->next = NULL;
-    (*info_p_p)->key = NULL;
-    (*info_p_p)->value = NULL;
+    info_init(*info_p_p);
 
   fn_fail:
     return mpi_errno;
@@ -70,7 +64,33 @@ int MPIR_Info_alloc(MPIR_Info ** info_p_p)
  * finalization by MPI_Info_create_env().  This routine must be thread-safe. */
 void MPIR_Info_setup_env(MPIR_Info * info_ptr)
 {
-    /* FIXME: Currently this info object is left empty, we need to add data to
-     * this as defined by the standard. */
-    (void) info_ptr;
+    info_init(info_ptr);
+    /* FIXME: Currently this info object is missing some data as
+     * defined by the standard. */
+}
+
+#define INFO_INITIAL_SIZE 10
+int MPIR_Info_push(MPIR_Info * info_ptr, const char *key, const char *val)
+{
+    int mpi_errno = MPI_SUCCESS;
+
+    /* potentially grow the copacity */
+    if (info_ptr->capacity == 0) {
+        info_ptr->entries = MPL_direct_malloc(sizeof(*(info_ptr->entries)) * INFO_INITIAL_SIZE);
+        info_ptr->capacity = INFO_INITIAL_SIZE;
+    } else if (info_ptr->size == info_ptr->capacity) {
+        int n = info_ptr->capacity * 5 / 3;     /* arbitrary grow ratio */
+        info_ptr->entries = MPL_direct_realloc(info_ptr->entries, sizeof(*(info_ptr->entries)) * n);
+        info_ptr->capacity = n;
+
+    }
+
+    /* add new entry */
+    int i = info_ptr->size;
+    info_ptr->entries[i].key = MPL_direct_strdup(key);
+    info_ptr->entries[i].value = MPL_direct_strdup(val);
+
+    info_ptr->size++;
+
+    return mpi_errno;
 }
