@@ -1,7 +1,6 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *   Copyright (C) 2008 University of Chicago.
- *   See COPYRIGHT notice in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
 
 #include "adio.h"
@@ -38,7 +37,7 @@ static void post_client_comm(ADIO_File fd, int rw_type,
  * - persistent file domains
  * - an option to use alltoall instead of point-to-point
  */
-void ADIOI_IOStridedColl(ADIO_File fd, void *buf, int count, int rdwr,
+void ADIOI_IOStridedColl(ADIO_File fd, void *buf, MPI_Aint count, int rdwr,
                          MPI_Datatype datatype, int file_ptr_type,
                          ADIO_Offset offset, ADIO_Status * status, int *error_code)
 {
@@ -50,7 +49,7 @@ void ADIOI_IOStridedColl(ADIO_File fd, void *buf, int count, int rdwr,
     int interleave_count = 0, i, nprocs, myrank, nprocs_for_coll;
     int cb_enable;
     ADIO_Offset bufsize;
-    MPI_Aint extent;
+    MPI_Aint lb, extent;
 #ifdef DEBUG2
     MPI_Aint bufextent;
 #endif
@@ -179,7 +178,7 @@ void ADIOI_IOStridedColl(ADIO_File fd, void *buf, int count, int rdwr,
         return;
     }
 
-    MPI_Type_extent(datatype, &extent);
+    MPI_Type_get_extent(datatype, &lb, &extent);
 #ifdef DEBUG2
     bufextent = extent * count;
 #endif
@@ -364,26 +363,38 @@ void ADIOI_IOStridedColl(ADIO_File fd, void *buf, int count, int rdwr,
                                          &agg_comm_requests, &aggs_client_count);
 
                 if (fd->is_agg && aggs_client_count) {
+#ifdef MPI_STATUSES_IGNORE
+                    agg_comm_statuses = MPI_STATUSES_IGNORE;
+#else
                     agg_comm_statuses = ADIOI_Malloc(aggs_client_count * sizeof(MPI_Status));
+#endif
                     MPI_Waitall(aggs_client_count, agg_comm_requests, agg_comm_statuses);
 #ifdef AGGREGATION_PROFILE
                     MPE_Log_event(5033, 0, NULL);
 #endif
                     ADIOI_Free(agg_comm_requests);
+#ifndef MPI_STATUSES_IGNORE
                     ADIOI_Free(agg_comm_statuses);
+#endif
                 }
 
                 if (clients_agg_count) {
+#ifdef MPI_STATUSES_IGNORE
+                    client_comm_statuses = MPI_STATUSES_IGNORE;
+#else
                     client_comm_statuses = ADIOI_Malloc(clients_agg_count * sizeof(MPI_Status));
+#endif
                     MPI_Waitall(clients_agg_count, client_comm_requests, client_comm_statuses);
 #ifdef AGGREGATION_PROFILE
                     MPE_Log_event(5039, 0, NULL);
 #endif
                     ADIOI_Free(client_comm_requests);
+#ifndef MPI_STATUSES_IGNORE
                     ADIOI_Free(client_comm_statuses);
+#endif
                 }
 #ifdef DEBUG2
-                fprintf(stderr, "buffered_io_size = %lld\n", buffered_io_size);
+                fprintf(stderr, "buffered_io_size = %lld\n", (long long) buffered_io_size);
                 if (fd->is_agg && buffered_io_size) {
                     fprintf(stderr, "buf = [");
                     for (i = 0; i < bufextent; i++)
@@ -423,17 +434,23 @@ void ADIOI_IOStridedColl(ADIO_File fd, void *buf, int count, int rdwr,
                 fflush(NULL);
 #endif
 #ifdef DEBUG
-                fprintf(stderr, "buffered_io_size = %lld\n", buffered_io_size);
+                fprintf(stderr, "buffered_io_size = %lld\n", (long long) buffered_io_size);
 #endif
 
                 if (clients_agg_count) {
+#ifdef MPI_STATUSES_IGNORE
+                    client_comm_statuses = MPI_STATUSES_IGNORE;
+#else
                     client_comm_statuses = ADIOI_Malloc(clients_agg_count * sizeof(MPI_Status));
+#endif
                     MPI_Waitall(clients_agg_count, client_comm_requests, client_comm_statuses);
 #ifdef AGGREGATION_PROFILE
                     MPE_Log_event(5039, 0, NULL);
 #endif
                     ADIOI_Free(client_comm_requests);
+#ifndef MPI_STATUSES_IGNORE
                     ADIOI_Free(client_comm_statuses);
+#endif
                 }
 #ifdef DEBUG2
                 if (bufextent) {
@@ -447,15 +464,21 @@ void ADIOI_IOStridedColl(ADIO_File fd, void *buf, int count, int rdwr,
                 if (fd->is_agg && buffered_io_size) {
                     ADIOI_Assert(aggs_client_count != 0);
                     /* make sure we actually have the data to write out */
+#ifdef MPI_STATUSES_IGNORE
+                    agg_comm_statuses = MPI_STATUSES_IGNORE;
+#else
                     agg_comm_statuses = (MPI_Status *)
                         ADIOI_Malloc(aggs_client_count * sizeof(MPI_Status));
+#endif
 
                     MPI_Waitall(aggs_client_count, agg_comm_requests, agg_comm_statuses);
 #ifdef AGGREGATION_PROFILE
                     MPE_Log_event(5033, 0, NULL);
 #endif
                     ADIOI_Free(agg_comm_requests);
+#ifndef MPI_STATUSES_IGNORE
                     ADIOI_Free(agg_comm_statuses);
+#endif
 #ifdef DEBUG2
                     fprintf(stderr, "cb_buf = [");
                     for (i = 0; i < buffered_io_size; i++)
@@ -640,13 +663,13 @@ void ADIOI_IOStridedColl(ADIO_File fd, void *buf, int count, int rdwr,
 
 /* Some of this code is from the old Calc_my_off_len() function.
  * It calculates the 1st and last byte accessed */
-void ADIOI_Calc_bounds(ADIO_File fd, int count, MPI_Datatype buftype,
+void ADIOI_Calc_bounds(ADIO_File fd, MPI_Aint count, MPI_Datatype buftype,
                        int file_ptr_type, ADIO_Offset offset,
                        ADIO_Offset * st_offset, ADIO_Offset * end_offset)
 {
     MPI_Count filetype_size, buftype_size, etype_size;
     int sum;
-    MPI_Aint filetype_extent;
+    MPI_Aint lb, filetype_extent;
     ADIO_Offset total_io;
     int filetype_is_contig;
     ADIO_Offset i, remainder;
@@ -671,7 +694,7 @@ void ADIOI_Calc_bounds(ADIO_File fd, int count, MPI_Datatype buftype,
 
     MPI_Type_size_x(fd->filetype, &filetype_size);
     ADIOI_Assert(filetype_size != 0);
-    MPI_Type_extent(fd->filetype, &filetype_extent);
+    MPI_Type_get_extent(fd->filetype, &lb, &filetype_extent);
     MPI_Type_size_x(fd->etype, &etype_size);
     MPI_Type_size_x(buftype, &buftype_size);
 
@@ -797,7 +820,8 @@ void ADIOI_Calc_bounds(ADIO_File fd, int count, MPI_Datatype buftype,
     *st_offset = st_byte_off;
     *end_offset = end_byte_off;
 #ifdef DEBUG
-    printf("st_offset = %lld\nend_offset = %lld\n", st_byte_off, end_byte_off);
+    printf("st_offset = %lld\nend_offset = %lld\n", (long long) st_byte_off,
+           (long long) end_byte_off);
 #endif
 #ifdef AGGREGATION_PROFILE
     MPE_Log_event(5001, 0, NULL);
@@ -809,7 +833,7 @@ void ADIOI_Calc_bounds(ADIO_File fd, int count, MPI_Datatype buftype,
  * WriteStrided call without affecting existing code.  For the new 2
  * phase code, we really only need to set a custom_ftype, and we can
  * assume that this uses MPI_BYTE for the etype, and disp is 0 */
-void ADIOI_IOFiletype(ADIO_File fd, void *buf, int count,
+void ADIOI_IOFiletype(ADIO_File fd, void *buf, MPI_Aint count,
                       MPI_Datatype datatype, int file_ptr_type,
                       ADIO_Offset offset, MPI_Datatype custom_ftype,
                       int rdwr, ADIO_Status * status, int *error_code)
@@ -821,7 +845,7 @@ void ADIOI_IOFiletype(ADIO_File fd, void *buf, int count,
     int user_ind_rd_buffer_size;
     int f_is_contig, m_is_contig;
     int user_ds_read, user_ds_write;
-    MPI_Aint f_extent;
+    MPI_Aint lb, f_extent;
     MPI_Count f_size;
     int f_ds_percent;           /* size/extent */
 
@@ -831,7 +855,7 @@ void ADIOI_IOFiletype(ADIO_File fd, void *buf, int count,
     else
         MPE_Log_event(5008, 0, NULL);
 #endif
-    MPI_Type_extent(custom_ftype, &f_extent);
+    MPI_Type_get_extent(custom_ftype, &lb, &f_extent);
     MPI_Type_size_x(custom_ftype, &f_size);
     f_ds_percent = 100 * f_size / f_extent;
 
@@ -871,7 +895,7 @@ void ADIOI_IOFiletype(ADIO_File fd, void *buf, int count,
     ADIOI_Datatype_iscontig(custom_ftype, &f_is_contig);
     ADIOI_Datatype_iscontig(datatype, &m_is_contig);
     if (!f_is_contig)
-        ADIOI_Flatten_datatype(custom_ftype);
+        ADIOI_Flatten_and_find(custom_ftype);
 
     /* make appropriate Read/Write calls.  Let ROMIO figure out file
      * system specific stuff. */
@@ -917,7 +941,6 @@ static void Exch_data_amounts(ADIO_File fd, int nprocs,
     MPI_Request *recv_requests;
     MPI_Request *send_requests;
     MPI_Status status;
-    MPI_Status *send_statuses;
     /* Aggregators send amounts for data requested to clients */
     if (fd->hints->cb_alltoall != ADIOI_HINT_DISABLE) {
         MPI_Alltoall(client_comm_sz_arr, sizeof(ADIO_Offset), MPI_BYTE,
@@ -984,10 +1007,14 @@ static void Exch_data_amounts(ADIO_File fd, int nprocs,
         ADIOI_Free(recv_requests);
         if (fd->is_agg) {
             /* wait for all sends to complete */
-            send_statuses = ADIOI_Malloc(nprocs * sizeof(MPI_Status));
+#ifdef MPI_STATUSES_IGNORE
+            MPI_Waitall(nprocs, send_requests, MPI_STATUSES_IGNORE);
+#else
+            MPI_Status *send_statuses = ADIOI_Malloc(nprocs * sizeof(MPI_Status));
             MPI_Waitall(nprocs, send_requests, send_statuses);
-            ADIOI_Free(send_requests);
             ADIOI_Free(send_statuses);
+#endif
+            ADIOI_Free(send_requests);
         }
     }
 }
