@@ -25,9 +25,9 @@ int MPIO_Err_create_code(int lastcode, int fatal, const char fcname[],
 
     buf = (char *) ADIOI_Malloc(1024);
     if (buf != NULL) {
-        idx += MPL_snprintf(buf, 1023, "%s (line %d): ", fcname, line);
+        idx += snprintf(buf, 1023, "%s (line %d): ", fcname, line);
         if (specific_msg == NULL) {
-            MPL_snprintf(&buf[idx], 1023 - idx, "%s\n", generic_msg);
+            snprintf(&buf[idx], 1023 - idx, "%s\n", generic_msg);
         } else {
             va_start(Argp, specific_msg);
             vsnprintf(&buf[idx], 1023 - idx, specific_msg, Argp);
