@@ -79,7 +79,6 @@ Input Parameters:
 int MPI_File_write_all_begin_c(MPI_File fh, ROMIO_CONST void *buf, MPI_Count count,
                                MPI_Datatype datatype)
 {
-    assert(count <= INT_MAX);
     int error_code;
     static char myname[] = "MPI_FILE_WRITE_ALL_BEGIN";
 
@@ -94,9 +93,8 @@ int MPI_File_write_all_begin_c(MPI_File fh, ROMIO_CONST void *buf, MPI_Count cou
 int MPIOI_File_write_all_begin(MPI_File fh,
                                MPI_Offset offset,
                                int file_ptr_type,
-                               const void *buf, int count, MPI_Datatype datatype, char *myname)
+                               const void *buf, MPI_Aint count, MPI_Datatype datatype, char *myname)
 {
-    assert(count <= INT_MAX);
     int error_code;
     MPI_Count datatype_size;
     ADIO_File adio_fh;
