@@ -8,9 +8,8 @@
 int MPIDI_CH3_Finalize(void)
 {
     int mpi_errno = MPI_SUCCESS;
-    MPIR_FUNC_VERBOSE_STATE_DECL(MPID_STATE_MPIDI_CH3_FINALIZE);
 
-    MPIR_FUNC_VERBOSE_ENTER(MPID_STATE_MPIDI_CH3_FINALIZE);
+    MPIR_FUNC_ENTER;
 
     mpi_errno = MPIDI_CH3I_Progress_finalize();
     MPIR_ERR_CHECK(mpi_errno);
@@ -19,6 +18,6 @@ int MPIDI_CH3_Finalize(void)
     if (mpi_errno) MPIR_ERR_POP (mpi_errno);
 
  fn_fail:
-    MPIR_FUNC_VERBOSE_EXIT(MPID_STATE_MPIDI_CH3_FINALIZE);
+    MPIR_FUNC_EXIT;
     return mpi_errno;
 }
