@@ -80,47 +80,44 @@ void MPID_Request_create_hook(MPIR_Request *req)
 int MPIDI_CH3U_Request_load_send_iov(MPIR_Request * const sreq,
 				     struct iovec * const iov, int * const iov_n)
 {
-    MPI_Aint last;
     int mpi_errno = MPI_SUCCESS;
 
     MPIR_FUNC_ENTER;
 
-    last = sreq->dev.msgsize;
-    MPL_DBG_MSG_FMT(MPIDI_CH3_DBG_CHANNEL,VERBOSE,(MPL_DBG_FDEST,
-     "pre-pv: first=%" PRIdPTR ", last=%" PRIdPTR ", iov_n=%d",
-		      sreq->dev.msg_offset, last, *iov_n));
-    MPIR_Assert(sreq->dev.msg_offset < last);
-    MPIR_Assert(last > 0);
-    MPIR_Assert(*iov_n > 0 && *iov_n <= MPL_IOV_LIMIT);
+    MPI_Aint density;
+    MPIR_Datatype_get_density(sreq->dev.datatype, density);
+    if (density >= MPIDI_IOV_DENSITY_MIN) {
+        MPI_Aint last = sreq->dev.msgsize;
+        MPL_DBG_MSG_FMT(MPIDI_CH3_DBG_CHANNEL,VERBOSE,(MPL_DBG_FDEST,
+        "pre-pv: first=%" PRIdPTR ", last=%" PRIdPTR ", iov_n=%d",
+                        sreq->dev.msg_offset, last, *iov_n));
+        MPIR_Assert(sreq->dev.msg_offset < last);
+        MPIR_Assert(last > 0);
+        MPIR_Assert(*iov_n > 0 && *iov_n <= MPL_IOV_LIMIT);
 
-    int max_iov_len = *iov_n;
-    MPI_Aint actual_iov_bytes, actual_iov_len;
-    MPIR_Typerep_to_iov(sreq->dev.user_buf, sreq->dev.user_count, sreq->dev.datatype,
-                     sreq->dev.msg_offset, iov, (MPI_Aint) max_iov_len,
-                     sreq->dev.msgsize - sreq->dev.msg_offset, &actual_iov_len, &actual_iov_bytes);
-    *iov_n = (int) actual_iov_len;
-    last = sreq->dev.msg_offset + actual_iov_bytes;
+        int max_iov_len = *iov_n;
+        MPI_Aint actual_iov_bytes, actual_iov_len;
+        MPIR_Typerep_to_iov(sreq->dev.user_buf, sreq->dev.user_count, sreq->dev.datatype,
+                        sreq->dev.msg_offset, iov, (MPI_Aint) max_iov_len,
+                        sreq->dev.msgsize - sreq->dev.msg_offset, &actual_iov_len, &actual_iov_bytes);
+        *iov_n = (int) actual_iov_len;
+        last = sreq->dev.msg_offset + actual_iov_bytes;
 
-    MPL_DBG_MSG_FMT(MPIDI_CH3_DBG_CHANNEL,VERBOSE,(MPL_DBG_FDEST,
-    "post-pv: first=%" PRIdPTR ", last=%" PRIdPTR ", iov_n=%d",
-		      sreq->dev.msg_offset, last, *iov_n));
-    MPIR_Assert(*iov_n > 0 && *iov_n <= MPL_IOV_LIMIT);
-    
-    if (last == sreq->dev.msgsize)
-    {
-	MPL_DBG_MSG(MPIDI_CH3_DBG_CHANNEL,VERBOSE,"remaining data loaded into IOV");
-	sreq->dev.OnDataAvail = sreq->dev.OnFinal;
-    }
-    else if ((last - sreq->dev.msg_offset) / *iov_n >= MPIDI_IOV_DENSITY_MIN)
-    {
-	MPL_DBG_MSG(MPIDI_CH3_DBG_CHANNEL,VERBOSE,"more data loaded into IOV");
-	sreq->dev.msg_offset = last;
-	sreq->dev.OnDataAvail = MPIDI_CH3_ReqHandler_SendReloadIOV;
-    }
-    else
-    {
-	intptr_t data_sz;
-	int i, iov_data_copied;
+        MPL_DBG_MSG_FMT(MPIDI_CH3_DBG_CHANNEL,VERBOSE,(MPL_DBG_FDEST,
+        "post-pv: first=%" PRIdPTR ", last=%" PRIdPTR ", iov_n=%d",
+                        sreq->dev.msg_offset, last, *iov_n));
+        MPIR_Assert(*iov_n > 0 && *iov_n <= MPL_IOV_LIMIT);
+        
+        if (last == sreq->dev.msgsize) {
+            MPL_DBG_MSG(MPIDI_CH3_DBG_CHANNEL,VERBOSE,"remaining data loaded into IOV");
+            sreq->dev.OnDataAvail = sreq->dev.OnFinal;
+        } else {
+            MPL_DBG_MSG(MPIDI_CH3_DBG_CHANNEL,VERBOSE,"more data loaded into IOV");
+            sreq->dev.msg_offset = last;
+            sreq->dev.OnDataAvail = MPIDI_CH3_ReqHandler_SendReloadIOV;
+        }
+    } else {
+	MPI_Aint data_sz;
 	
 	MPL_DBG_MSG(MPIDI_CH3_DBG_CHANNEL,VERBOSE,"low density.  using SRBuf.");
 	    
@@ -141,29 +138,21 @@ int MPIDI_CH3U_Request_load_send_iov(MPIR_Request * const sreq,
 	    /* --END ERROR HANDLING-- */
 	}
 
-	iov_data_copied = 0;
-	for (i = 0; i < *iov_n; i++) {
-	    MPIR_Memcpy((char*) sreq->dev.tmpbuf + iov_data_copied,
-		   iov[i].iov_base, iov[i].iov_len);
-	    iov_data_copied += iov[i].iov_len;
-	}
-	sreq->dev.msg_offset = last;
-
         MPI_Aint max_pack_bytes;
         MPI_Aint actual_pack_bytes;
 
-        if (data_sz > sreq->dev.tmpbuf_sz - iov_data_copied)
-            max_pack_bytes = sreq->dev.tmpbuf_sz - iov_data_copied;
+        if (data_sz > sreq->dev.tmpbuf_sz)
+            max_pack_bytes = sreq->dev.tmpbuf_sz;
         else
-            max_pack_bytes = sreq->dev.msgsize - sreq->dev.msg_offset;
+            max_pack_bytes = data_sz;
 
         MPIR_Typerep_pack(sreq->dev.user_buf, sreq->dev.user_count, sreq->dev.datatype,
-                       sreq->dev.msg_offset, (char*) sreq->dev.tmpbuf + iov_data_copied,
-                       max_pack_bytes, &actual_pack_bytes);
-        last = sreq->dev.msg_offset + actual_pack_bytes;
+                       sreq->dev.msg_offset, (char*) sreq->dev.tmpbuf,
+                       max_pack_bytes, &actual_pack_bytes, MPIR_TYPEREP_FLAG_NONE);
+        MPI_Aint last = sreq->dev.msg_offset + actual_pack_bytes;
 
 	iov[0].iov_base = (void *)sreq->dev.tmpbuf;
-	iov[0].iov_len = actual_pack_bytes + iov_data_copied;
+	iov[0].iov_len = actual_pack_bytes;
 	*iov_n = 1;
 	if (last == sreq->dev.msgsize)
 	{
@@ -418,7 +407,7 @@ int MPIDI_CH3U_Request_unpack_srbuf(MPIR_Request * rreq)
     MPI_Aint actual_unpack_bytes;
     MPIR_Typerep_unpack(rreq->dev.tmpbuf, tmpbuf_last - rreq->dev.msg_offset,
                      rreq->dev.user_buf, rreq->dev.user_count, rreq->dev.datatype,
-                     rreq->dev.msg_offset, &actual_unpack_bytes);
+                     rreq->dev.msg_offset, &actual_unpack_bytes, MPIR_TYPEREP_FLAG_NONE);
     last = rreq->dev.msg_offset + actual_unpack_bytes;
 
     if (last == 0 || last == rreq->dev.msg_offset)
@@ -528,7 +517,7 @@ int MPIDI_CH3U_Request_unpack_uebuf(MPIR_Request * rreq)
 	    MPI_Aint actual_unpack_bytes;
 	    MPIR_Typerep_unpack(rreq->dev.tmpbuf, unpack_sz,
 			     rreq->dev.user_buf, rreq->dev.user_count,
-			     rreq->dev.datatype, 0, &actual_unpack_bytes);
+			     rreq->dev.datatype, 0, &actual_unpack_bytes, MPIR_TYPEREP_FLAG_NONE);
 
 	    if (actual_unpack_bytes != unpack_sz)
 	    {
@@ -556,10 +545,6 @@ int MPID_Request_complete(MPIR_Request *req)
 
     MPIDI_CH3U_Request_decrement_cc(req, &incomplete);
     if (!incomplete) {
-        /* decrement completion_notification counter */
-        if (req->completion_notification)
-            MPIR_cc_dec(req->completion_notification);
-
 	MPIR_Request_free(req);
     }
 
