@@ -5,10 +5,11 @@
 
 #include "mpidimpl.h"
 
-int MPID_Probe(int source, int tag, MPIR_Comm * comm, int context_offset,
+int MPID_Probe(int source, int tag, MPIR_Comm * comm, int attr,
 	       MPI_Status * status)
 {
     MPID_Progress_state progress_state;
+    int context_offset = MPIR_PT2PT_ATTR_CONTEXT_OFFSET(attr);
     const int context = comm->recvcontext_id + context_offset;
     int mpi_errno = MPI_SUCCESS;
 
@@ -30,9 +31,7 @@ int MPID_Probe(int source, int tag, MPIR_Comm * comm, int context_offset,
             do {
                 int found;
                 
-                MPID_THREAD_CS_ENTER(POBJ, MPIR_THREAD_POBJ_MSGQ_MUTEX);
                 found = MPIDI_CH3U_Recvq_FU(source, tag, context, status);
-                MPID_THREAD_CS_EXIT(POBJ, MPIR_THREAD_POBJ_MSGQ_MUTEX);
                 if (found) goto fn_exit;
 
                 mpi_errno = MPIDI_Anysource_iprobe_fn(tag, comm, context_offset, &found, status);
@@ -74,9 +73,7 @@ int MPID_Probe(int source, int tag, MPIR_Comm * comm, int context_offset,
     {
         int found;
 
-        MPID_THREAD_CS_ENTER(POBJ, MPIR_THREAD_POBJ_MSGQ_MUTEX);
         found = MPIDI_CH3U_Recvq_FU(source, tag, context, status);
-        MPID_THREAD_CS_EXIT(POBJ, MPIR_THREAD_POBJ_MSGQ_MUTEX);
         if (found) break;
 
 	mpi_errno = MPIDI_CH3_Progress_wait(&progress_state);
