@@ -36,6 +36,7 @@
     (sreq_)->dev.user_count = count;					\
     (sreq_)->dev.datatype = datatype;					\
     (sreq_)->u.persist.real_request = NULL;                             \
+    MPIR_Comm_save_inactive_request(comm, sreq_);                       \
 }
 
 	
@@ -54,16 +55,20 @@ int MPID_Startall(int count, MPIR_Request * requests[])
     {
 	MPIR_Request * const preq = requests[i];
 
-        /* continue if the source/dest is MPI_PROC_NULL */
-        if (preq->dev.match.parts.rank == MPI_PROC_NULL)
-            continue;
-
         if (preq->kind == MPIR_REQUEST_KIND__PREQUEST_COLL) {
             mpi_errno = MPIR_Persist_coll_start(preq);
             MPIR_ERR_CHECK(mpi_errno);
             continue;
         }
 
+        /* only pt2pt requests should reach here */
+        MPIR_Assert(preq->kind == MPIR_REQUEST_KIND__PREQUEST_SEND ||
+                    preq->kind == MPIR_REQUEST_KIND__PREQUEST_RECV);
+
+        /* continue if the source/dest is MPI_PROC_NULL */
+        if (preq->dev.match.parts.rank == MPI_PROC_NULL)
+            continue;
+
 	/* FIXME: The odd 7th arg (match.context_id - comm->context_id) 
 	   is probably to get the context offset.  Do we really need the
 	   context offset? Is there any case where the offset isn't zero? */
@@ -161,7 +166,7 @@ int MPID_Startall(int count, MPIR_Request * requests[])
 /*
  * MPID_Send_init()
  */
-int MPID_Send_init(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPIR_Comm * comm, int context_offset,
+int MPID_Send_init(const void * buf, MPI_Aint count, MPI_Datatype datatype, int rank, int tag, MPIR_Comm * comm, int attr,
 		   MPIR_Request ** request)
 {
     MPIR_Request * sreq;
@@ -169,6 +174,7 @@ int MPID_Send_init(const void * buf, int count, MPI_Datatype datatype, int rank,
 
     MPIR_FUNC_ENTER;
 
+    int context_offset = MPIR_PT2PT_ATTR_CONTEXT_OFFSET(attr);
     MPIDI_Request_create_psreq(sreq, mpi_errno, goto fn_exit);
     MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_SEND);
     if (!HANDLE_IS_BUILTIN(datatype))
@@ -186,7 +192,7 @@ int MPID_Send_init(const void * buf, int count, MPI_Datatype datatype, int rank,
 /*
  * MPID_Ssend_init()
  */
-int MPID_Ssend_init(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPIR_Comm * comm, int context_offset,
+int MPID_Ssend_init(const void * buf, MPI_Aint count, MPI_Datatype datatype, int rank, int tag, MPIR_Comm * comm, int attr,
 		    MPIR_Request ** request)
 {
     MPIR_Request * sreq;
@@ -194,13 +200,10 @@ int MPID_Ssend_init(const void * buf, int count, MPI_Datatype datatype, int rank
 
     MPIR_FUNC_ENTER;
 
+    int context_offset = MPIR_PT2PT_ATTR_CONTEXT_OFFSET(attr);
     MPIDI_Request_create_psreq(sreq, mpi_errno, goto fn_exit);
     MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_SSEND);
-    if (!HANDLE_IS_BUILTIN(datatype))
-    {
-	MPIR_Datatype_get_ptr(datatype, sreq->dev.datatype_ptr);
-    MPIR_Datatype_ptr_add_ref(sreq->dev.datatype_ptr);
-    }
+    MPIR_Datatype_add_ref_if_not_builtin(datatype);
     *request = sreq;
 
   fn_exit:    
@@ -211,7 +214,7 @@ int MPID_Ssend_init(const void * buf, int count, MPI_Datatype datatype, int rank
 /*
  * MPID_Rsend_init()
  */
-int MPID_Rsend_init(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPIR_Comm * comm, int context_offset,
+int MPID_Rsend_init(const void * buf, MPI_Aint count, MPI_Datatype datatype, int rank, int tag, MPIR_Comm * comm, int attr,
 		    MPIR_Request ** request)
 {
     MPIR_Request * sreq;
@@ -219,6 +222,7 @@ int MPID_Rsend_init(const void * buf, int count, MPI_Datatype datatype, int rank
 
     MPIR_FUNC_ENTER;
 
+    int context_offset = MPIR_PT2PT_ATTR_CONTEXT_OFFSET(attr);
     MPIDI_Request_create_psreq(sreq, mpi_errno, goto fn_exit);
     MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_RSEND);
     if (!HANDLE_IS_BUILTIN(datatype))
@@ -236,7 +240,7 @@ int MPID_Rsend_init(const void * buf, int count, MPI_Datatype datatype, int rank
 /*
  * MPID_Bsend_init()
  */
-int MPID_Bsend_init(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPIR_Comm * comm, int context_offset,
+int MPID_Bsend_init(const void * buf, MPI_Aint count, MPI_Datatype datatype, int rank, int tag, MPIR_Comm * comm, int attr,
 		    MPIR_Request ** request)
 {
     MPIR_Request * sreq;
@@ -244,6 +248,7 @@ int MPID_Bsend_init(const void * buf, int count, MPI_Datatype datatype, int rank
 
     MPIR_FUNC_ENTER;
 
+    int context_offset = MPIR_PT2PT_ATTR_CONTEXT_OFFSET(attr);
     MPIDI_Request_create_psreq(sreq, mpi_errno, goto fn_exit);
     MPIDI_Request_set_type(sreq, MPIDI_REQUEST_TYPE_BSEND);
     if (!HANDLE_IS_BUILTIN(datatype))
@@ -269,7 +274,7 @@ int MPID_Bsend_init(const void * buf, int count, MPI_Datatype datatype, int rank
 /*
  * MPID_Recv_init()
  */
-int MPID_Recv_init(void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPIR_Comm * comm, int context_offset,
+int MPID_Recv_init(void * buf, MPI_Aint count, MPI_Datatype datatype, int rank, int tag, MPIR_Comm * comm, int context_offset,
 		   MPIR_Request ** request)
 {
     MPIR_Request * rreq;
@@ -297,6 +302,7 @@ int MPID_Recv_init(void * buf, int count, MPI_Datatype datatype, int rank, int t
     rreq->dev.user_count = count;
     rreq->dev.datatype = datatype;
     rreq->u.persist.real_request = NULL;
+    MPIR_Comm_save_inactive_request(comm, rreq);
     MPIDI_Request_set_type(rreq, MPIDI_REQUEST_TYPE_RECV);
     if (!HANDLE_IS_BUILTIN(datatype))
     {
