@@ -121,7 +121,7 @@ static int win_allgather(MPIR_Win * win, void *base, int disp_unit)
 {
     int i, same_disp, mpi_errno = MPI_SUCCESS;
     uint32_t first;
-    MPIR_Errflag_t errflag = MPIR_ERR_NONE;
+    uint64_t local_key = 0;
     MPIR_Comm *comm_ptr = win->comm_ptr;
     MPIDI_OFI_win_targetinfo_t *winfo;
     int nic = 0;
@@ -129,25 +129,53 @@ static int win_allgather(MPIR_Win * win, void *base, int disp_unit)
     MPIR_FUNC_ENTER;
 
     if (!MPIDI_OFI_ENABLE_MR_PROV_KEY) {
-        MPIDI_OFI_WIN(win).mr_key = MPIDI_OFI_WIN(win).win_id;
+        if (MPIDIG_WIN(win, info_args).optimized_mr &&
+            MPIDIG_WIN(win, info_args).accumulate_ordering == 0) {
+            /* accumulate_ordering must be set to zero to support creation of optimized
+             * memory region for use with lower-overhead, unordered RMA operations.
+             * Attempting to create an optimized memory region key. Gets the next MR key that's
+             * available to the processes involved in the RMA window. Use the current maximum + 1
+             * to ensure that the key is available for all processes. */
+            mpi_errno = MPIR_Allreduce(&MPIDI_OFI_global.global_max_optimized_mr_key, &local_key, 1,
+                                       MPI_UNSIGNED, MPI_MAX, comm_ptr, MPIR_ERR_NONE);
+            MPIR_ERR_CHECK(mpi_errno);
+
+            if (local_key + 1 < MPIDI_OFI_NUM_OPTIMIZED_MEMORY_REGIONS) {
+                MPIDI_OFI_global.global_max_optimized_mr_key = local_key + 1;
+                MPIDI_OFI_WIN(win).mr_key = local_key;
+            }
+        }
+        /* Assign regular memory registration key if the optimized one is
+         * not requested or exhausted */
+        if (local_key + 1 >= MPIDI_OFI_NUM_OPTIMIZED_MEMORY_REGIONS ||
+            !MPIDIG_WIN(win, info_args).optimized_mr) {
+            /* Makes sure that regular mr key does not fall within optimized mr key range */
+            MPIDI_OFI_WIN(win).mr_key =
+                MPIDI_OFI_NUM_OPTIMIZED_MEMORY_REGIONS + MPIDI_OFI_WIN(win).win_id;
+        }
     } else {
+        /* Expect provider to assign the key */
         MPIDI_OFI_WIN(win).mr_key = 0;
     }
 
-    /* we need register mr on the correct domain for the vni */
-    int vni = MPIDI_OFI_get_win_vni(win);
-    int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, vni, nic);
+    /* we need register mr on the correct domain for the vci */
+    int vci = MPIDI_WIN(win, am_vci);
+    int ctx_idx = MPIDI_OFI_get_ctx_index(vci, nic);
 
     /* Register the allocated win buffer or MPI_BOTTOM (NULL) for dynamic win.
      * It is clear that we cannot register NULL when FI_MR_ALLOCATED is set, thus
      * we skip trial and return immediately. When FI_MR_ALLOCATED is not set, however,
      * it lacks documentation defining whether the provider can accept NULL and whether
      * it means registration of the entire virtual address space, although we have observed
-     * successful use in existing providers. Thus, we take a try here. */
-
+     * successful use in most of the existing providers. Thus, we take a try here for most
+     * providers. For providers like CXI, FI_MR_ALLOCATED is not set but registration with
+     * non-zero address is not supported. For such providers, registration is skipped by
+     * using the MPIDI_OFI_ENABLE_MR_REGISTER_NULL capability set variable. */
     int rc = 0, allrc = 0;
     MPIDI_OFI_WIN(win).mr = NULL;
-    if (base || (win->create_flavor == MPI_WIN_FLAVOR_DYNAMIC && !MPIDI_OFI_ENABLE_MR_ALLOCATED)) {
+
+    if (base || (win->create_flavor == MPI_WIN_FLAVOR_DYNAMIC &&
+                 !MPIDI_OFI_ENABLE_MR_ALLOCATED && MPIDI_OFI_ENABLE_MR_REGISTER_NULL)) {
         size_t len;
         if (win->create_flavor == MPI_WIN_FLAVOR_DYNAMIC) {
             len = UINTPTR_MAX - (uintptr_t) base;
@@ -155,10 +183,19 @@ static int win_allgather(MPIR_Win * win, void *base, int disp_unit)
             len = (size_t) win->size;
         }
 
-        /* device buffers are not currently supported */
-        if (MPIR_GPU_query_pointer_is_dev(base))
-            rc = -1;
-        else
+        if (MPIR_GPU_query_pointer_is_strict_dev(base, NULL)) {
+            if (MPIDI_OFI_ENABLE_HMEM) {
+                MPL_pointer_attr_t attr;
+                MPIR_GPU_query_pointer_attr(base, &attr);
+                mpi_errno =
+                    MPIDI_OFI_register_memory(base, len, &attr, ctx_idx, MPIDI_OFI_WIN(win).mr_key,
+                                              &MPIDI_OFI_WIN(win).mr);
+                if (mpi_errno != MPI_SUCCESS)
+                    rc = -1;
+            } else {
+                rc = -1;
+            }
+        } else {
             MPIDI_OFI_CALL_RETURN(fi_mr_reg(MPIDI_OFI_global.ctx[ctx_idx].domain,       /* In:  Domain Object */
                                             base,       /* In:  Lower memory address */
                                             len,        /* In:  Length              */
@@ -168,16 +205,22 @@ static int win_allgather(MPIR_Win * win, void *base, int disp_unit)
                                             0ULL,       /* In:  flags               */
                                             &MPIDI_OFI_WIN(win).mr,     /* Out: memregion object    */
                                             NULL), rc); /* In:  context             */
+        }
+        if (rc == 0) {
+            mpi_errno = MPIDI_OFI_mr_bind(MPIDI_OFI_global.prov_use[0], MPIDI_OFI_WIN(win).mr,
+                                          MPIDI_OFI_WIN(win).ep, MPIDI_OFI_WIN(win).cmpl_cntr);
+            MPIR_ERR_CHECK(mpi_errno);
+        }
     } else if (win->create_flavor == MPI_WIN_FLAVOR_DYNAMIC) {
         /* We may still do native atomics with collective attach, let's load acc_hint */
         load_acc_hint(win);
         goto fn_exit;
     } else {
-        /* FIXME: what's in the branch? If it can't happen, add an assertion here */
+        /* Do nothing */
     }
 
     /* Check if any process fails to register. If so, release local MR and force AM path. */
-    MPIR_Allreduce(&rc, &allrc, 1, MPI_INT, MPI_MIN, comm_ptr, &errflag);
+    MPIR_Allreduce(&rc, &allrc, 1, MPI_INT, MPI_MIN, comm_ptr, MPIR_ERR_NONE);
     if (allrc < 0) {
         if (rc >= 0 && MPIDI_OFI_WIN(win).mr)
             MPIDI_OFI_CALL(fi_close(&MPIDI_OFI_WIN(win).mr->fid), fi_close);
@@ -201,7 +244,7 @@ static int win_allgather(MPIR_Win * win, void *base, int disp_unit)
 
     mpi_errno = MPIR_Allgather(MPI_IN_PLACE, 0,
                                MPI_DATATYPE_NULL,
-                               winfo, sizeof(*winfo), MPI_BYTE, comm_ptr, &errflag);
+                               winfo, sizeof(*winfo), MPI_BYTE, comm_ptr, MPIR_ERR_NONE);
     MPIR_ERR_CHECK(mpi_errno);
 
     if (!MPIDI_OFI_ENABLE_MR_PROV_KEY && !MPIDI_OFI_ENABLE_MR_VIRT_ADDRESS) {
@@ -241,7 +284,8 @@ static int win_set_per_win_sync(MPIR_Win * win)
 {
     int ret, mpi_errno = MPI_SUCCESS;
     int nic = 0;
-    int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, MPIDI_OFI_WIN(win).vni, nic);
+    int vci = MPIDI_WIN(win, am_vci);
+    int ctx_idx = MPIDI_OFI_get_ctx_index(vci, nic);
 
     MPIR_FUNC_ENTER;
 
@@ -281,7 +325,7 @@ static int win_set_per_win_sync(MPIR_Win * win)
 
 static void win_init_am(MPIR_Win * win)
 {
-    MPIDI_WIN(win, am_vci) %= MPIDI_OFI_global.num_vnis;
+    MPIR_Assert(MPIDI_WIN(win, am_vci) < MPIDI_OFI_global.num_vcis);
 }
 
 /*
@@ -298,8 +342,8 @@ static int win_init_sep(MPIR_Win * win)
     int i, ret, mpi_errno = MPI_SUCCESS;
     struct fi_info *finfo;
     int nic = 0;
-    int vni = MPIDI_OFI_WIN(win).vni;
-    int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, vni, nic);
+    int vci = MPIDI_WIN(win, am_vci);
+    int ctx_idx = MPIDI_OFI_get_ctx_index(vci, nic);
 
     MPIR_FUNC_ENTER;
 
@@ -309,7 +353,7 @@ static int win_init_sep(MPIR_Win * win)
 
     /* Initialize scalable EP when first window is created. */
     if (MPIDI_OFI_global.ctx[ctx_idx].rma_sep == NULL) {
-        /* NOTE: if MPIDI_OFI_VNI_USE_SEPCTX, we could share rma_stx_ctx across vnis */
+        /* NOTE: if MPIDI_OFI_VNI_USE_SEPCTX, we could share rma_stx_ctx across vcis */
 
         MPIDI_OFI_global.max_rma_sep_tx_cnt =
             MPL_MIN(MPIDI_OFI_global.prov_use[0]->domain_attr->max_ep_tx_ctx,
@@ -426,8 +470,8 @@ static int win_init_stx(MPIR_Win * win)
     struct fi_info *finfo;
     bool have_per_win_cntr = false;
     int nic = 0;
-    int vni = MPIDI_OFI_WIN(win).vni;
-    int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, vni, nic);
+    int vci = MPIDI_WIN(win, am_vci);
+    int ctx_idx = MPIDI_OFI_get_ctx_index(vci, nic);
 
     MPIR_FUNC_ENTER;
 
@@ -438,7 +482,7 @@ static int win_init_stx(MPIR_Win * win)
 
     /* Initialize rma shared context when first window is created. */
     if (MPIDI_OFI_global.ctx[ctx_idx].rma_stx_ctx == NULL) {
-        /* NOTE: if MPIDI_OFI_VNI_USE_SEPCTX, we could share rma_stx_ctx across vnis */
+        /* NOTE: if MPIDI_OFI_VNI_USE_SEPCTX, we could share rma_stx_ctx across vcis */
 
         struct fi_tx_attr tx_attr;
         memset(&tx_attr, 0, sizeof(tx_attr));
@@ -540,16 +584,16 @@ static int win_init_global(MPIR_Win * win)
 {
     MPIR_FUNC_ENTER;
 
-    int vni = MPIDI_OFI_WIN(win).vni;
+    int vci = MPIDI_WIN(win, am_vci);
     int nic = 0;
-    int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, vni, nic);
+    int ctx_idx = MPIDI_OFI_get_ctx_index(vci, nic);
 
     MPIDI_OFI_WIN(win).ep = MPIDI_OFI_global.ctx[ctx_idx].tx;
 #ifdef MPIDI_OFI_VNI_USE_DOMAIN
     MPIDI_OFI_WIN(win).cmpl_cntr = MPIDI_OFI_global.ctx[ctx_idx].rma_cmpl_cntr;
     MPIDI_OFI_WIN(win).issued_cntr = &MPIDI_OFI_global.ctx[ctx_idx].rma_issued_cntr;
 #else
-    ctx_idx = MPIDI_OFI_get_ctx_index(NULL, 0, nic);
+    ctx_idx = MPIDI_OFI_get_ctx_index(0, nic);
     /* NOTE: shared with ctx[0] */
     MPIDI_OFI_WIN(win).cmpl_cntr = MPIDI_OFI_global.ctx[ctx_idx].rma_cmpl_cntr;
     MPIDI_OFI_WIN(win).issued_cntr = &MPIDI_OFI_global.ctx[ctx_idx].rma_issued_cntr;
@@ -572,6 +616,9 @@ static int win_init(MPIR_Win * win)
 
     MPIDI_OFI_WIN(win).win_id =
         MPIDI_OFI_mr_key_alloc(MPIDI_OFI_COLL_MR_KEY, win->comm_ptr->context_id);
+    MPIR_ERR_CHKANDJUMP(MPIDI_OFI_WIN(win).win_id == MPIDI_OFI_INVALID_MR_KEY, mpi_errno,
+                        MPI_ERR_OTHER, "**ofid_mr_key");
+
     if (MPIDI_OFI_WIN(win).win_id == -1ULL) {
         MPL_DBG_MSG(MPIDI_CH4_DBG_GENERAL, VERBOSE, "Failed to get global mr key.\n");
         mpi_errno = MPIDI_OFI_ENAVAIL;
@@ -583,12 +630,6 @@ static int win_init(MPIR_Win * win)
     MPIDI_OFI_WIN(win).sep_tx_idx = -1; /* By default, -1 means not using scalable EP. */
     MPIDI_OFI_WIN(win).progress_counter = 1;
 
-    /* Assign vni to window.
-     * NOTE: we could assign vni per epoch, then we need run `win_init_{sep,stx,global}`
-     * at start of every epoch.
-     */
-    MPIDI_OFI_WIN(win).vni = MPIDI_OFI_get_win_vni(win);
-
     /* First, try to enable scalable EP. */
     if (MPIR_CVAR_CH4_OFI_ENABLE_SCALABLE_ENDPOINTS && MPIR_CVAR_CH4_OFI_MAX_RMA_SEP_CTX > 0) {
         /* Create tx based on scalable EP. */
@@ -609,6 +650,8 @@ static int win_init(MPIR_Win * win)
   fn_exit:
     MPIR_FUNC_EXIT;
     return mpi_errno;
+  fn_fail:
+    goto fn_exit;
 }
 
 /* Callback of MPL_gavl_tree_create to delete local registered MR for dynamic window */
@@ -884,12 +927,12 @@ int MPIDI_OFI_mpi_win_attach_hook(MPIR_Win * win, void *base, MPI_Aint size)
     int mpi_errno = MPI_SUCCESS;
     MPIR_FUNC_ENTER;
 
-    MPIR_Errflag_t errflag = MPIR_ERR_NONE;
     MPIR_Comm *comm_ptr = win->comm_ptr;
     dwin_target_mr_t *target_mrs;
     int mpl_err = MPL_SUCCESS, i;
     int nic = 0;
-    int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, MPIDI_OFI_WIN(win).vni, nic);
+    int vci = MPIDI_WIN(win, am_vci);
+    int ctx_idx = MPIDI_OFI_get_ctx_index(vci, nic);
 
     MPIR_FUNC_ENTER;
 
@@ -899,15 +942,18 @@ int MPIDI_OFI_mpi_win_attach_hook(MPIR_Win * win, void *base, MPI_Aint size)
         goto fn_exit;
 
     uint64_t requested_key = 0ULL;
-    if (!MPIDI_OFI_ENABLE_MR_PROV_KEY)
+    if (!MPIDI_OFI_ENABLE_MR_PROV_KEY) {
         requested_key = MPIDI_OFI_mr_key_alloc(MPIDI_OFI_LOCAL_MR_KEY, MPIDI_OFI_INVALID_MR_KEY);
+        MPIR_ERR_CHKANDJUMP(requested_key == MPIDI_OFI_INVALID_MR_KEY, mpi_errno,
+                            MPI_ERR_OTHER, "**ofid_mr_key");
+    }
 
     int rc = 0, allrc = 0;
     struct fid_mr *mr = NULL;
     /* device buffers are not currently supported */
     if (MPIR_GPU_query_pointer_is_dev(base))
         rc = -1;
-    else
+    else {
         MPIDI_OFI_CALL_RETURN(fi_mr_reg(MPIDI_OFI_global.ctx[ctx_idx].domain,   /* In:  Domain Object */
                                         base,   /* In:  Lower memory address */
                                         size,   /* In:  Length              */
@@ -917,9 +963,13 @@ int MPIDI_OFI_mpi_win_attach_hook(MPIR_Win * win, void *base, MPI_Aint size)
                                         0ULL,   /* In:  flags               */
                                         &mr,    /* Out: memregion object    */
                                         NULL), rc);     /* In:  context             */
+        mpi_errno = MPIDI_OFI_mr_bind(MPIDI_OFI_global.prov_use[0], mr,
+                                      MPIDI_OFI_WIN(win).ep, MPIDI_OFI_WIN(win).cmpl_cntr);
+        MPIR_ERR_CHECK(mpi_errno);
+    }
 
     /* Check if any process fails to register. If so, release local MR and force AM path. */
-    MPIR_Allreduce(&rc, &allrc, 1, MPI_INT, MPI_MIN, comm_ptr, &errflag);
+    MPIR_Allreduce(&rc, &allrc, 1, MPI_INT, MPI_MIN, comm_ptr, MPIR_ERR_NONE);
     if (allrc < 0) {
         if (rc >= 0)
             MPIDI_OFI_CALL(fi_close(&mr->fid), fi_close);
@@ -944,7 +994,8 @@ int MPIDI_OFI_mpi_win_attach_hook(MPIR_Win * win, void *base, MPI_Aint size)
     target_mrs[comm_ptr->rank].size = (uintptr_t) size;
     mpi_errno = MPIR_Allgather(MPI_IN_PLACE, 0,
                                MPI_DATATYPE_NULL,
-                               target_mrs, sizeof(dwin_target_mr_t), MPI_BYTE, comm_ptr, &errflag);
+                               target_mrs, sizeof(dwin_target_mr_t), MPI_BYTE, comm_ptr,
+                               MPIR_ERR_NONE);
     MPIR_ERR_CHECK(mpi_errno);
 
     /* Insert each remote MR which will be searched when issuing an RMA operation
@@ -976,7 +1027,6 @@ int MPIDI_OFI_mpi_win_detach_hook(MPIR_Win * win, const void *base)
     int mpi_errno = MPI_SUCCESS;
     MPIR_FUNC_ENTER;
 
-    MPIR_Errflag_t errflag = MPIR_ERR_NONE;
     MPIR_Comm *comm_ptr = win->comm_ptr;
     const void **target_bases;
     int mpl_err = MPL_SUCCESS, i;
@@ -1002,7 +1052,8 @@ int MPIDI_OFI_mpi_win_detach_hook(MPIR_Win * win, const void *base)
      * that all processes collectively call detach. */
     target_bases[comm_ptr->rank] = base;
     mpi_errno = MPIR_Allgather(MPI_IN_PLACE, 0, MPI_DATATYPE_NULL,
-                               target_bases, sizeof(const void *), MPI_BYTE, comm_ptr, &errflag);
+                               target_bases, sizeof(const void *), MPI_BYTE, comm_ptr,
+                               MPIR_ERR_NONE);
     MPIR_ERR_CHECK(mpi_errno);
 
     /* Search and delete each remote MR */
@@ -1029,8 +1080,8 @@ int MPIDI_OFI_mpi_win_free_hook(MPIR_Win * win)
     MPIR_FUNC_ENTER;
 
     if (MPIDI_OFI_ENABLE_RMA) {
-        int vni = MPIDI_OFI_WIN(win).vni;
-        int ctx_idx = MPIDI_OFI_get_ctx_index(NULL, vni, nic);
+        int vci = MPIDI_WIN(win, am_vci);
+        int ctx_idx = MPIDI_OFI_get_ctx_index(vci, nic);
         MPIDI_OFI_mr_key_free(MPIDI_OFI_COLL_MR_KEY, MPIDI_OFI_WIN(win).win_id);
         MPIDIU_map_erase(MPIDI_OFI_global.win_map, MPIDI_OFI_WIN(win).win_id);
         /* For scalable EP: push transmit context index back into available pool. */
