@@ -13,7 +13,13 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_SHM_mpi_improbe(int source,
                                                    int *flag, MPIR_Request ** message,
                                                    MPI_Status * status)
 {
-    return MPIDIG_mpi_improbe(source, tag, comm, context_offset, 0, flag, message, status);
+    int mpi_errno = MPI_SUCCESS;
+
+    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(0).lock);
+    mpi_errno = MPIDIG_mpi_improbe(source, tag, comm, context_offset, 0, flag, message, status);
+    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(0).lock);
+
+    return mpi_errno;
 }
 
 MPL_STATIC_INLINE_PREFIX int MPIDI_SHM_mpi_iprobe(int source,
@@ -22,7 +28,13 @@ MPL_STATIC_INLINE_PREFIX int MPIDI_SHM_mpi_iprobe(int source,
                                                   int context_offset,
                                                   int *flag, MPI_Status * status)
 {
-    return MPIDIG_mpi_iprobe(source, tag, comm, context_offset, 0, flag, status);
+    int mpi_errno = MPI_SUCCESS;
+
+    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(0).lock);
+    mpi_errno = MPIDIG_mpi_iprobe(source, tag, comm, context_offset, 0, flag, status);
+    MPID_THREAD_CS_ENTER(VCI, MPIDI_VCI(0).lock);
+
+    return mpi_errno;
 }
 
 #endif /* SHM_AM_FALLBACK_PROBE_H_INCLUDED */
