@@ -1,3 +1,8 @@
+/*
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
+ */
+
 #include <mpi.h>
 #include <stdio.h>
 #include "mpitest.h"
@@ -14,7 +19,7 @@ int main(int argc, char **argv)
     comm = MPI_COMM_WORLD;
     MPI_Comm_group(comm, &basegroup);
     MPI_Comm_rank(comm, &rank);
-    MPI_Errhandler_set(MPI_COMM_WORLD, MPI_ERRORS_RETURN);
+    MPI_Comm_set_errhandler(MPI_COMM_WORLD, MPI_ERRORS_RETURN);
 
     /* Checking group_rank for NULL variable */
     mpi_errno = MPI_Group_rank(basegroup, NULL);
@@ -22,7 +27,6 @@ int main(int argc, char **argv)
     if (errclass != MPI_ERR_ARG)
         ++errs;
 
-    MPI_Comm_free(&comm);
     MPI_Group_free(&basegroup);
     MTest_Finalize(errs);
     return 0;
