@@ -8,7 +8,7 @@
 
 int main(int argc, char **argv)
 {
-    int rank, size;
+    int rank;
 
     MPI_Init(&argc, &argv);
     MPI_Comm_rank(MPI_COMM_WORLD, &rank);
