@@ -1,9 +1,8 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *
- *  (C) 2003 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
+
 #include <mpi.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -15,7 +14,7 @@
  */
 int main(int argc, char **argv)
 {
-    int rank, err;
+    int rank, err = 0;
     char buf[10];
 
     MPI_Init(&argc, &argv);
