@@ -1,14 +1,13 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *
- *  (C) 2003 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
+
 #include <stdio.h>
 #include "mpi.h"
 #include "mpitest.h"
 
-/* Test bsend with a buffer with arbitray alignment */
+/* Test bsend with a buffer with arbitrary alignment */
 #define BUFSIZE 2000*4
 int main(int argc, char *argv[])
 {
@@ -62,7 +61,7 @@ int main(int argc, char *argv[])
         if (bptr != buf + align) {
             errs++;
             printf
-                ("Did not recieve the same buffer on detach that was provided on init (%p vs %p)\n",
+                ("Did not receive the same buffer on detach that was provided on init (%p vs %p)\n",
                  bptr, buf);
         }
     }
