@@ -1,9 +1,8 @@
-/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
 /*
- *
- *  (C) 2001 by Argonne National Laboratory.
- *      See COPYRIGHT in top-level directory.
+ * Copyright (C) by Argonne National Laboratory
+ *     See COPYRIGHT in top-level directory
  */
+
 #include "mpi.h"
 #include <stdio.h>
 #include "mpitest.h"
@@ -33,7 +32,8 @@ int main(int argc, char *argv[])
 
         /* Create key values */
         for (i = 0; i < 3; i++) {
-            MPI_Win_create_keyval(MPI_NULL_COPY_FN, MPI_NULL_DELETE_FN, &key[i], (void *) 0);
+            MPI_Win_create_keyval(MPI_WIN_NULL_COPY_FN, MPI_WIN_NULL_DELETE_FN, &key[i],
+                                  (void *) 0);
             attrval[i] = 1024 * i;
         }
 
@@ -98,7 +98,7 @@ int checkAttrs(MPI_Win win, int n, int key[], int attrval[])
             fprintf(stderr, "Attribute for key %d not set\n", i);
         } else if (val_p != &attrval[i]) {
             errs++;
-            fprintf(stderr, "Atribute value for key %d not correct\n", i);
+            fprintf(stderr, "Attribute value for key %d not correct\n", i);
         }
     }
 
