@@ -125,7 +125,7 @@ int main(int argc, char *argv[])
         /* Perform several communication operations, mixing synchronization
          * types.  Use multiple communication to avoid the single-operation
          * optimization that may be present. */
-        MTestPrintfMsg(3, "Begining loop %d of mixed sync put/acc operations\n", loop);
+        MTestPrintfMsg(3, "Beginning loop %d of mixed sync put/acc operations\n", loop);
         memset(winbuf, 0, count * sizeof(int));
         MPI_Barrier(comm);
         if (crank == source) {
@@ -170,7 +170,7 @@ int main(int argc, char *argv[])
         /* Perform several communication operations, mixing synchronization
          * types.  Use multiple communication to avoid the single-operation
          * optimization that may be present. */
-        MTestPrintfMsg(3, "Begining loop %d of mixed sync put/get/acc operations\n", loop);
+        MTestPrintfMsg(3, "Beginning loop %d of mixed sync put/get/acc operations\n", loop);
         MPI_Barrier(comm);
         if (crank == source) {
             MPI_Win_lock(MPI_LOCK_EXCLUSIVE, dest, 0, win);
